from enum import Enum
from typing import Literal

from pydantic import BaseConfig, BaseModel, Extra

from kpops.utils.pydantic import CamelCaseConfig


class KafkaConnectorType(str, Enum):
    SINK = "sink"
    SOURCE = "source"


class KafkaConnectConfig(BaseModel):
    class Config(BaseConfig):
        extra = Extra.allow


class ConnectorTask(BaseModel):
    connector: str
    task: int


class KafkaConnectResponse(BaseModel):
    name: str
    config: dict[str, str]
    tasks: list[ConnectorTask]
    type: str | None

    class Config(BaseConfig):
        extra = Extra.forbid


class KafkaConnectConfigError(BaseModel):
    name: str
    errors: list[str]


class KafkaConnectConfigDescription(BaseModel):
    value: KafkaConnectConfigError


class KafkaConnectConfigErrorResponse(BaseModel):
    name: str
    error_count: int
    configs: list[KafkaConnectConfigDescription]


class KafkaConnectResetterConfig(BaseModel):
    brokers: str
    connector: str
    delete_consumer_group: bool | None = None
    offset_topic: str | None = None

    class Config(CamelCaseConfig):
        pass


class KafkaConnectResetterValues(BaseModel):
    connector_type: Literal["source", "sink"]
    config: KafkaConnectResetterConfig
    name_override: str

    class Config(CamelCaseConfig):
        pass

    def dict(self, **_) -> dict:
        return super().dict(by_alias=True, exclude_none=True)



import logging

from kpops.component_handlers.topic.exception import TopicNotFoundException
from kpops.component_handlers.topic.model import (
    TopicConfigResponse,
    TopicResponse,
    TopicSpec,
)
from kpops.component_handlers.topic.proxy_wrapper import HEADERS, ProxyWrapper
from kpops.component_handlers.topic.utils import (
    get_effective_config,
    parse_and_compare_topic_configs,
    parse_rest_proxy_topic_config,
)
from kpops.components.base_components.models.to_section import TopicConfig, ToSection
from kpops.utils.colorify import greenify, magentaify, yellowify
from kpops.utils.dict_differ import Diff, DiffType, get_diff, render_diff

log = logging.getLogger("KafkaTopic")


class TopicHandler:
    def __init__(self, proxy_wrapper: ProxyWrapper):
        self.proxy_wrapper = proxy_wrapper

    def create_topics(self, to_section: ToSection, dry_run: bool) -> None:
        topics: dict[str, TopicConfig] = to_section.topics
        for topic_name, topic_config in topics.items():
            topic_spec = self.__prepare_body(topic_name, topic_config)
            if dry_run:
                self.__dry_run_topic_creation(topic_name, topic_spec, topic_config)
            else:
                try:
                    self.proxy_wrapper.get_topic(topic_name=topic_name)
                    topic_config_in_cluster = self.proxy_wrapper.get_topic_config(
                        topic_name=topic_name
                    )
                    differences = self.__get_topic_config_diff(
                        topic_config_in_cluster, topic_config.configs
                    )

                    if differences:
                        json_body = []
                        for difference in differences:
                            if difference.diff_type == DiffType.REMOVE:
                                json_body.append(
                                    {"name": difference.key, "operation": "DELETE"}
                                )
                            else:
                                config_value = difference.change.new_value
                                if config_value:
                                    json_body.append(
                                        {"name": difference.key, "value": config_value}
                                    )
                        self.proxy_wrapper.batch_alter_topic_config(
                            topic_name=topic_name,
                            json_body=json_body,
                        )

                    else:
                        log.info(
                            f"Topic Creation: config of topic {topic_name} didn't change. Skipping update."
                        )
                except TopicNotFoundException:
                    self.proxy_wrapper.create_topic(topic_spec=topic_spec)

    def delete_topics(self, to_section: ToSection, dry_run: bool) -> None:
        topics: dict[str, TopicConfig] = to_section.topics
        for topic_name in topics.keys():
            if dry_run:
                self.__dry_run_topic_deletion(topic_name=topic_name)
            else:
                try:
                    self.proxy_wrapper.get_topic(topic_name=topic_name)
                    self.proxy_wrapper.delete_topic(topic_name=topic_name)
                except TopicNotFoundException:
                    log.warning(
                        f"Topic Deletion: topic {topic_name} does not exist in the cluster and cannot be deleted. Skipping."
                    )

    @staticmethod
    def __get_topic_config_diff(
        cluster_config: TopicConfigResponse, current_config: dict
    ) -> list[Diff]:
        comparable_in_cluster_config_dict, _ = parse_rest_proxy_topic_config(
            cluster_config
        )

        return get_diff(comparable_in_cluster_config_dict, current_config)

    def __dry_run_topic_creation(
        self,
        topic_name: str,
        topic_spec: TopicSpec,
        topic_config: TopicConfig | None = None,
    ) -> None:
        try:
            topic_in_cluster = self.proxy_wrapper.get_topic(topic_name=topic_name)
            topic_name = topic_in_cluster.topic_name
            if topic_config:
                topic_config_in_cluster = self.proxy_wrapper.get_topic_config(
                    topic_name=topic_name
                )
                in_cluster_config, new_config = parse_and_compare_topic_configs(
                    topic_config_in_cluster, topic_config.configs
                )
                if diff := render_diff(in_cluster_config, new_config):
                    log.info(f"Config changes for topic {topic_name}:")
                    log.info("\n" + diff)

            log.info(f"Topic Creation: {topic_name} already exists in cluster.")
            log.debug("HTTP/1.1 400 Bad Request")
            log.debug(HEADERS)
            error_message = {
                "error_code": 40002,
                "message": f"Topic '{topic_name}' already exists.",
            }
            log.debug(error_message)

            broker_config = self.proxy_wrapper.get_broker_config()
            effective_config = get_effective_config(broker_config)

            self.__check_partition_count(topic_in_cluster, topic_spec, effective_config)
            self.__check_replication_factor(
                topic_in_cluster, topic_spec, effective_config
            )
        except TopicNotFoundException:
            log.info(
                greenify(
                    f"Topic Creation: {topic_name} does not exist in the cluster. Creating topic."
                )
            )
            log.debug(f"POST /clusters/{self.proxy_wrapper.cluster_id}/topics HTTP/1.1")
            log.debug(f"Host: {self.proxy_wrapper.host}")
            log.debug(HEADERS)
            log.debug(topic_spec.dict())

    @staticmethod
    def __check_partition_count(
        topic_in_cluster: TopicResponse,
        topic_spec: TopicSpec,
        broker_config: dict[str, str],
    ) -> None:
        topic_name = topic_in_cluster.topic_name
        partition_count = topic_in_cluster.partitions_count
        if partition_count == (
            topic_spec.partitions_count or int(broker_config["num.partitions"])
        ):
            log.info(
                yellowify(
                    f"Topic Creation: partition count of topic {topic_name} did not change. Current partitions count {partition_count}. Updating configs."
                )
            )
        else:
            log.error(
                f"Topic Creation: partition count of topic {topic_name} changed! Partitions count of topic {topic_name} is {partition_count}. The given partitions count {topic_spec.partitions_count}."
            )
            exit(1)

    @staticmethod
    def __check_replication_factor(
        topic_in_cluster: TopicResponse,
        topic_spec: TopicSpec,
        broker_config: dict[str, str],
    ) -> None:
        topic_name = topic_in_cluster.topic_name
        replication_factor = topic_in_cluster.replication_factor
        if replication_factor == (
            topic_spec.replication_factor
            or int(broker_config["default.replication.factor"])
        ):
            log.info(
                yellowify(
                    f"Topic Creation: replication factor of topic {topic_name} did not change. Current replication factor {replication_factor}. Updating configs."
                )
            )
        else:
            log.error(
                f"Topic Creation: replication factor of topic {topic_name} changed! Replication factor of topic {topic_name} is {replication_factor}. The given replication count {topic_spec.replication_factor}."
            )
            exit(1)

    def __dry_run_topic_deletion(self, topic_name: str) -> None:
        try:
            topic_in_cluster = self.proxy_wrapper.get_topic(topic_name=topic_name)
            log.info(
                magentaify(
                    f"Topic Deletion: topic {topic_in_cluster.topic_name} exists in the cluster. Deleting topic."
                )
            )
            log.debug(
                f"DELETE /clusters/{self.proxy_wrapper.cluster_id}/topics HTTP/1.1"
            )
        except TopicNotFoundException:
            log.warning(
                f"Topic Deletion: topic {topic_name} does not exist in the cluster and cannot be deleted. Skipping."
            )
            log.debug(f"Host: {self.proxy_wrapper.host}")
            log.debug(HEADERS)
            log.debug("HTTP/1.1 404 Not Found")
            log.debug(HEADERS)
            error_message = {
                "error_code": 40403,
                "message": f"This server does not host the topic-partition '{topic_name}'.",
            }
            log.debug(error_message)

    @classmethod
    def __prepare_body(cls, topic_name: str, topic_config: TopicConfig) -> TopicSpec:
        """
        Prepares the POST request body needed for the topic creation
        :param topic_name: The name of the topic
        :param topic_config: The topic config
        :return:
        """
        topic_spec_json: dict = topic_config.dict(
            include={
                "partitions_count": True,
                "replication_factor": True,
                "configs": True,
            },
            exclude_unset=True,
            exclude_none=True,
        )
        configs = []
        for config_name, config_value in topic_spec_json["configs"].items():
            configs.append({"name": config_name, "value": config_value})
        topic_spec_json["configs"] = configs
        topic_spec_json["topic_name"] = topic_name
        return TopicSpec(**topic_spec_json)



from enum import Enum

from pydantic import BaseConfig, BaseModel, Extra, Field, root_validator


class InputTopicTypes(str, Enum):
    INPUT = "input"
    EXTRA = "extra"
    INPUT_PATTERN = "input-pattern"
    EXTRA_PATTERN = "extra-pattern"


class FromTopic(BaseModel):
    type: InputTopicTypes = Field(...)
    role: str | None = None

    class Config(BaseConfig):
        extra = Extra.forbid
        use_enum_values = True

    @root_validator
    def extra_topic_role(cls, values):
        is_extra_topic = values["type"] in (
            InputTopicTypes.EXTRA,
            InputTopicTypes.EXTRA_PATTERN,
        )
        if is_extra_topic and not values.get("role"):
            raise ValueError(
                "If you define an extra input topic or extra input pattern, you have to define a role."
            )
        if not is_extra_topic and values.get("role"):
            raise ValueError(
                "If you do not define an input topic or input pattern, the role is unnecessary."
            )
        return values


class FromSection(BaseModel):
    topics: dict[str, FromTopic]

    class Config(BaseConfig):
        extra = Extra.forbid



from kpops.component_handlers.topic.model import (
    BrokerConfigResponse,
    KafkaTopicConfigSource,
    TopicConfigResponse,
)


def parse_and_compare_topic_configs(
    topic_config_in_cluster: TopicConfigResponse, topic_config: dict
) -> tuple[dict, dict]:

    comparable_in_cluster_config_dict, default_configs = parse_rest_proxy_topic_config(
        topic_config_in_cluster
    )

    cluster_defaults_overwrite = set(topic_config.keys()) - set(
        comparable_in_cluster_config_dict.keys()
    )
    config_overwrites = set(comparable_in_cluster_config_dict.keys()) - set(
        topic_config.keys()
    )
    populate_default_configs(
        cluster_defaults_overwrite,
        default_configs,
        topic_config,
        description_text="was default",
    )

    populate_default_configs(
        config_overwrites,
        default_configs,
        comparable_in_cluster_config_dict,
        description_text="fallback to default",
    )

    return comparable_in_cluster_config_dict, topic_config


def populate_default_configs(
    config_overwrites: set,
    default_configs: dict,
    config_to_populate: dict,
    description_text: str,
):
    for overwrite in config_overwrites:
        default_config_value = default_configs.get(overwrite)
        if default_config_value:
            config_to_populate[overwrite] = (
                str(config_to_populate[overwrite])
                + f" ({description_text} {default_config_value})"
            )


def parse_rest_proxy_topic_config(
    topic_config_in_cluster: TopicConfigResponse,
) -> tuple[dict, dict]:
    comparable_in_cluster_config_dict = {}
    default_configs = {}
    for config in topic_config_in_cluster.data:
        if config.source == KafkaTopicConfigSource.DYNAMIC_TOPIC_CONFIG:
            comparable_in_cluster_config_dict[config.name] = config.value
        for synonym in config.synonyms:
            if synonym.source == KafkaTopicConfigSource.DEFAULT_CONFIG:
                default_configs[config.name] = synonym.value
                continue  # we only expect one default value here
    return comparable_in_cluster_config_dict, default_configs


def get_effective_config(
    broker_config_in_cluster: BrokerConfigResponse,
) -> dict[str, str]:
    return {
        config.name: config.value
        for config in broker_config_in_cluster.data
        if config.value
    }



import logging

import typer


class CustomFormatter(logging.Formatter):
    def format(self, record):
        message_format = "%(name)s - %(message)s"

        if record.name == "root":
            message_format = "%(message)s"

        formats = {
            logging.DEBUG: message_format,
            logging.INFO: message_format,
            logging.WARNING: typer.style(message_format, fg=typer.colors.YELLOW),
            logging.ERROR: typer.style(message_format, fg=typer.colors.RED),
            logging.CRITICAL: typer.style(
                message_format, fg=typer.colors.RED, bold=True
            ),
        }

        log_fmt = formats.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)



from pydantic import BaseConfig, BaseModel, Extra

from kpops.components.base_components.base_defaults_component import deduplicate
from kpops.components.base_components.kafka_app import (
    KafkaAppConfig,
    KafkaStreamsConfig,
)


class StreamsConfig(KafkaStreamsConfig):
    """
    Streams Bootstrap streams section
    """

    input_topics: list[str] = []
    input_pattern: str | None = None
    extra_input_topics: dict[str, list[str]] = {}
    extra_input_patterns: dict[str, str] = {}
    extra_output_topics: dict[str, str] = {}
    output_topic: str | None = None
    error_topic: str | None = None
    config: dict[str, str] = {}

    def add_input_topics(self, topics: list[str]) -> None:
        self.input_topics = deduplicate(self.input_topics + topics)

    def add_extra_input_topics(self, role: str, topics: list[str]) -> None:
        self.extra_input_topics[role] = deduplicate(
            self.extra_input_topics.get(role, []) + topics
        )

    def dict(
        self,
        *,
        include=None,
        exclude=None,
        by_alias: bool = False,
        skip_defaults: bool | None = None,
        exclude_unset: bool = False,
        **kwargs,
    ) -> dict:
        return super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            # The following lines are required only for the streams configs since we never not want to export defaults here, just fallback to helm default values
            exclude_defaults=True,
            exclude_none=True,
        )


class StreamsAppAutoScaling(BaseModel):
    consumergroup: str
    topics: list[str] = []

    class Config(BaseConfig):
        extra = Extra.allow


class StreamsAppConfig(KafkaAppConfig):
    """
    StreamsBoostrap app configurations.

    The attributes correspond to keys and values that are used as values for the streams bootstrap helm chart.
    """

    streams: StreamsConfig
    autoscaling: StreamsAppAutoScaling | None = None

    class Config(BaseConfig):
        extra = Extra.allow



from __future__ import annotations

import json
import logging
from functools import cached_property

from schema_registry.client import SchemaRegistryClient

from kpops.cli.exception import ClassNotFoundError
from kpops.cli.pipeline_config import PipelineConfig
from kpops.cli.registry import find_class
from kpops.component_handlers.schema_handler.schema_provider import (
    Schema,
    SchemaProvider,
)
from kpops.components.base_components.models.to_section import ToSection
from kpops.utils.colorify import greenify, magentaify, yellowify

log = logging.getLogger("SchemaHandler")


class SchemaHandler:
    def __init__(self, url: str, components_module: str | None):
        self.schema_registry_client = SchemaRegistryClient(url)
        self.components_module = components_module

    @cached_property
    def schema_provider(self) -> SchemaProvider:
        try:
            if not self.components_module:
                raise ValueError(
                    f"The Schema Registry URL is set but you haven't specified the component module path. Please provide a valid component module path where your {SchemaProvider.__name__} implementation exists."
                )
            schema_provider_class = find_class(self.components_module, SchemaProvider)  # type: ignore[type-abstract]
            return schema_provider_class()
        except ClassNotFoundError:
            raise ValueError(
                f"No schema provider found in components module {self.components_module}. "
                f"Please implement the abstract method in {SchemaProvider.__module__}.{SchemaProvider.__name__}."
            )

    @classmethod
    def load_schema_handler(
        cls, components_module: str | None, config: PipelineConfig
    ) -> SchemaHandler | None:
        if not config.schema_registry_url:
            return None

        return cls(
            url=config.schema_registry_url,
            components_module=components_module,
        )

    def submit_schemas(self, to_section: ToSection, dry_run: bool = True) -> None:
        for topic_name, config in to_section.topics.items():
            value_schema_class = config.value_schema
            key_schema_class = config.key_schema
            if value_schema_class is not None:
                schema = self.schema_provider.provide_schema(
                    value_schema_class, to_section.models
                )
                self.__submit_value_schema(
                    schema, value_schema_class, dry_run, topic_name
                )
            if key_schema_class is not None:
                schema = self.schema_provider.provide_schema(
                    key_schema_class, to_section.models
                )
                self.__submit_key_schema(schema, key_schema_class, dry_run, topic_name)

    def delete_schemas(self, to_section: ToSection, dry_run: bool = True) -> None:
        for topic_name, config in to_section.topics.items():
            if config.value_schema is not None:
                self.__delete_subject(f"{topic_name}-value", dry_run)
            if config.key_schema is not None:
                self.__delete_subject(f"{topic_name}-key", dry_run)

    def __submit_key_schema(
        self,
        schema: Schema,
        schema_class: str,
        dry_run: bool,
        topic_name: str,
    ) -> None:
        subject = f"{topic_name}-key"
        self.__submit_schema(
            subject=subject,
            schema=schema,
            schema_class=schema_class,
            dry_run=dry_run,
        )

    def __submit_value_schema(
        self,
        schema: Schema,
        schema_class: str,
        dry_run: bool,
        topic_name: str,
    ) -> None:
        subject = f"{topic_name}-value"
        self.__submit_schema(
            subject=subject,
            schema=schema,
            schema_class=schema_class,
            dry_run=dry_run,
        )

    def __submit_schema(
        self,
        subject: str,
        schema: Schema,
        schema_class: str,
        dry_run: bool,
    ):
        if dry_run:
            if self.__subject_exists(subject):
                self.__check_compatibility(schema, schema_class, subject)
            else:
                log.info(
                    greenify(
                        f"Schema Submission: The subject {subject} will be submitted."
                    )
                )
        else:
            self.schema_registry_client.register(subject=subject, schema=schema)
            log.info(
                f"Schema Submission: schema submitted for {subject} with model {schema_class}."
            )

    def __subject_exists(self, subject: str) -> bool:
        return len(self.schema_registry_client.get_versions(subject)) > 0

    def __check_compatibility(
        self, schema: Schema, schema_class: str, subject: str
    ) -> None:
        registered_version = self.schema_registry_client.check_version(subject, schema)
        if registered_version is None:
            if not self.schema_registry_client.test_compatibility(
                subject=subject, schema=schema
            ):
                raise Exception(
                    f"Schema is not compatible for {subject} and model {schema_class}. \n {json.dumps(schema.flat_schema, indent=4)}"
                )
        else:
            log.info(
                yellowify(
                    f"Schema Submission: schema was already submitted for the subject {subject} as version {registered_version.schema}. Therefore, the specified schema must be compatible."
                )
            )

        log.info(
            f"Schema Submission: compatible schema for {subject} with model {schema_class}."
        )

    def __delete_subject(self, subject: str, dry_run: bool) -> None:
        if dry_run:
            log.info(magentaify(f"Schema Deletion: will delete subject {subject}."))
        else:
            version_list = self.schema_registry_client.delete_subject(subject)
            log.info(
                f"Schema Deletion: deleted {len(version_list)} versions for subject {subject}."
            )



import logging

log = logging.getLogger("HelmUtils")


RELEASE_NAME_MAX_LEN = 52


def trim_release_name(name: str, suffix: str = "") -> str:
    """
    Trim Helm release name while preserving suffix.
    :param name: The release name including optional suffix
    :param suffix: The release suffix to preserve
    :return: Truncated release name
    """
    if len(name) > RELEASE_NAME_MAX_LEN:
        new_name = name[: (RELEASE_NAME_MAX_LEN - len(suffix))] + suffix
        log.critical(
            f"Invalid Helm release name '{name}'. Truncating to {RELEASE_NAME_MAX_LEN} characters: \n {name} --> {new_name}"
        )
        name = new_name
    return name



import logging
from functools import cached_property

import requests

from kpops.cli.pipeline_config import PipelineConfig
from kpops.component_handlers.topic.exception import (
    KafkaRestProxyError,
    TopicNotFoundException,
)
from kpops.component_handlers.topic.model import (
    BrokerConfigResponse,
    TopicConfigResponse,
    TopicResponse,
    TopicSpec,
)

log = logging.getLogger("KafkaRestProxy")

HEADERS = {"Content-Type": "application/json"}


class ProxyWrapper:
    """
    Wraps Kafka REST Proxy APIs
    """

    def __init__(self, pipeline_config: PipelineConfig) -> None:
        if not pipeline_config.kafka_rest_host:
            raise ValueError(
                "The Kafka REST Proxy host is not set. Please set the host in the config.yaml using the kafka_rest_host property or set the environemt variable KPOPS_REST_PROXY_HOST."
            )

        self._host = pipeline_config.kafka_rest_host

    @cached_property
    def cluster_id(self) -> str:
        """
        Gets the Kafka cluster ID by sending a requests to Kafka REST proxy.
        More information about the cluster ID can be found here:
        https://docs.confluent.io/platform/current/kafka-rest/api.html#cluster-v3

        Currently both Kafka and Kafka REST Proxy are only aware of the Kafka cluster pointed at by the
        bootstrap.servers configuration. Therefore, only one Kafka cluster will be returned.
        :return: The Kafka cluster ID.
        """
        response = requests.get(url=f"{self._host}/v3/clusters")
        if response.status_code == requests.status_codes.codes.ok:
            cluster_information = response.json()
            return cluster_information["data"][0]["cluster_id"]

        raise KafkaRestProxyError(response)

    @property
    def host(self) -> str:
        return self._host

    def create_topic(self, topic_spec: TopicSpec) -> None:
        """
        Creates a topic.
        API Reference: https://docs.confluent.io/platform/current/kafka-rest/api.html#post--clusters-cluster_id-topics
        :param topic_spec: The topic specification.
        """
        response = requests.post(
            url=f"{self._host}/v3/clusters/{self.cluster_id}/topics",
            headers=HEADERS,
            json=topic_spec.dict(exclude_unset=True, exclude_none=True),
        )
        if response.status_code == requests.status_codes.codes.created:
            log.info(f"Topic {topic_spec.topic_name} created.")
            log.debug(response.json())
            return

        raise KafkaRestProxyError(response)

    def delete_topic(self, topic_name: str) -> None:
        """
        Deletes a topic
        API Reference: https://docs.confluent.io/platform/current/kafka-rest/api.html#delete--clusters-cluster_id-topics-topic_name
        :param topic_name: Name of the topic
        """
        response = requests.delete(
            url=f"{self.host}/v3/clusters/{self.cluster_id}/topics/{topic_name}",
            headers=HEADERS,
        )
        if response.status_code == requests.status_codes.codes.no_content:
            log.info(f"Topic {topic_name} deleted.")
            return

        raise KafkaRestProxyError(response)

    def get_topic(self, topic_name: str) -> TopicResponse:
        """
        Returns the topic with the given topic_name.
        API Reference: https://docs.confluent.io/platform/current/kafka-rest/api.html#get--clusters-cluster_id-topics-topic_name
        :param topic_name: The topic name.
        :return: Response of the get topic API
        """
        response = requests.get(
            url=f"{self.host}/v3/clusters/{self.cluster_id}/topics/{topic_name}",
            headers=HEADERS,
        )
        if response.status_code == requests.status_codes.codes.ok:
            log.debug(f"Topic {topic_name} found.")
            log.debug(response.json())
            return TopicResponse(**response.json())

        elif (
            response.status_code == requests.status_codes.codes.not_found
            and response.json()["error_code"] == 40403
        ):
            log.debug(f"Topic {topic_name} not found.")
            log.debug(response.json())
            raise TopicNotFoundException()

        raise KafkaRestProxyError(response)

    def get_topic_config(self, topic_name: str) -> TopicConfigResponse:
        """
        Return the config with the given topic_name.
        API Reference: https://docs.confluent.io/platform/current/kafka-rest/api.html#acl-v3
        :param topic_name: The topic name.
        :return: The topic configuration.
        """
        response = requests.get(
            url=f"{self.host}/v3/clusters/{self.cluster_id}/topics/{topic_name}/configs",
            headers=HEADERS,
        )

        if response.status_code == requests.status_codes.codes.ok:
            log.debug(f"Configs for {topic_name} found.")
            log.debug(response.json())
            return TopicConfigResponse(**response.json())

        elif (
            response.status_code == requests.status_codes.codes.not_found
            and response.json()["error_code"] == 40403
        ):
            log.debug(f"Configs for {topic_name} not found.")
            log.debug(response.json())
            raise TopicNotFoundException()

        raise KafkaRestProxyError(response)

    def batch_alter_topic_config(self, topic_name: str, json_body: list[dict]) -> None:
        """
        Reset config of given config_name param to the default value on the kafka server.
        API Reference: https://docs.confluent.io/platform/current/kafka-rest/api.html#post--clusters-cluster_id-topics-topic_name-configs-alter
        :param topic_name: The topic name.
        :param config_name: The configuration parameter name.
        """
        response = requests.post(
            url=f"{self.host}/v3/clusters/{self.cluster_id}/topics/{topic_name}/configs:alter",
            headers=HEADERS,
            json={"data": json_body},
        )
        if response.status_code == requests.status_codes.codes.no_content:
            log.info(f"Config of topic {topic_name} was altered.")
            return

        raise KafkaRestProxyError(response)

    def get_broker_config(self) -> BrokerConfigResponse:
        """
        Return the list of configuration parameters for all the brokers in the given Kafka cluster.
        API Reference: https://docs.confluent.io/platform/current/kafka-rest/api.html#get--clusters-cluster_id-brokers---configs
        :return: The broker configuration.
        """
        response = requests.get(
            url=f"{self.host}/v3/clusters/{self.cluster_id}/brokers/-/configs",
            headers=HEADERS,
        )

        if response.status_code == requests.status_codes.codes.ok:
            log.debug("Broker configs found.")
            log.debug(response.json())
            return BrokerConfigResponse(**response.json())

        raise KafkaRestProxyError(response)



import inspect
import logging
import os
from collections import deque
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import TypeVar

import typer
from pydantic import BaseConfig, BaseModel, Field

from kpops.cli.pipeline_config import PipelineConfig
from kpops.component_handlers import ComponentHandlers
from kpops.utils.yaml_loading import load_yaml_file

log = logging.getLogger("PipelineComponentEnricher")


class BaseDefaultsComponent(BaseModel):
    _type: str = Field(..., alias="type")

    enrich: bool = Field(default=False, exclude=True)
    config: PipelineConfig = Field(default=..., exclude=True)
    handlers: ComponentHandlers = Field(default=..., exclude=True)

    class Config(BaseConfig):
        arbitrary_types_allowed = True

    def __init__(self, **kwargs):
        if kwargs.get("enrich", True):
            kwargs = self.extend_with_defaults(kwargs)
        super().__init__(**kwargs)

    def extend_with_defaults(self, kwargs) -> dict:
        """
        Merges tmp_defaults with all tmp_defaults for parent classes

        :param kwargs: the init kwargs for pydantic
        :return: enriched kwargs with tmp_defaults
        """

        config: PipelineConfig = kwargs["config"]
        log.debug(
            typer.style(
                "Enriching component of type ", fg=typer.colors.GREEN, bold=False
            )
            + typer.style(
                kwargs.get("type"), fg=typer.colors.GREEN, bold=True, underline=True
            )
        )
        # Merge tmp_defaults before we initialize the thing
        classes = deque(inspect.getmro(self.__class__))
        classes.appendleft(self.__class__)
        for base in deduplicate(classes):
            if issubclass(base, BaseDefaultsComponent):
                (
                    main_default_file_path,
                    environment_default_file_path,
                ) = get_defaults_file_paths(config)
                if not environment_default_file_path.exists():
                    kwargs = update_nested(
                        kwargs,
                        defaults_from_yaml(main_default_file_path, base._type),
                    )
                else:
                    kwargs = update_nested(
                        kwargs,
                        defaults_from_yaml(environment_default_file_path, base._type),
                        defaults_from_yaml(main_default_file_path, base._type),
                    )
        return kwargs


def defaults_from_yaml(path: Path, key: str) -> dict:
    """
    Read value from json file and return @default if not found
    """
    content = load_yaml_file(path, substitution=dict(os.environ))
    if not isinstance(content, dict):
        raise TypeError(
            "Default files should be structured as map ([app type] -> [default config]"
        )
    value = content.get(key)
    if value is None:
        return {}
    log.debug(
        f"\tFound defaults for component type {typer.style(key, bold=True, fg=typer.colors.MAGENTA)} in file:  {path}"
    )
    return value


def get_defaults_file_paths(config: PipelineConfig) -> tuple[Path, Path]:
    defaults_dir = Path(config.defaults_path).resolve()
    main_default_file_path = defaults_dir / Path(
        config.defaults_filename_prefix
    ).with_suffix(".yaml")

    environment_default_file_path = defaults_dir / Path(
        f"{config.defaults_filename_prefix}_{config.environment}"
    ).with_suffix(".yaml")

    return main_default_file_path, environment_default_file_path


T = TypeVar("T")


def deduplicate(seq: Sequence[T]) -> list[T]:
    """Deduplicate items of a sequence while preserving its order."""
    return list(dict.fromkeys(seq))


def update_nested_pair(original_dict: dict, other_dict: Mapping) -> dict:
    """Nested update for 2 dictionaries"""
    for key, value in other_dict.items():
        if isinstance(value, Mapping):
            nested_val = original_dict.get(key, {})
            if nested_val is not None:
                original_dict[key] = update_nested_pair(nested_val, value)
        else:
            if key not in original_dict:
                original_dict[key] = value
    return original_dict


def update_nested(*argv: dict) -> dict:
    """
    Merge multiple configuration dicts. The dicts have multiple layers. These layers will be merged recursively.
    :arg argv - n dictionaries
    """

    if len(argv) == 0:
        return {}
    if len(argv) == 1:
        return argv[0]
    if len(argv) == 2:
        return update_nested_pair(argv[0], argv[1])
    return update_nested(update_nested_pair(argv[0], argv[1]), *argv[2:])



from pydantic import BaseConfig, Extra
from typing_extensions import override

from kpops.component_handlers.helm_wrapper.model import HelmRepoConfig
from kpops.components.base_components.kafka_app import KafkaApp
from kpops.components.base_components.models.to_section import (
    OutputTopicTypes,
    TopicConfig,
)
from kpops.components.streams_bootstrap.app_type import AppType
from kpops.components.streams_bootstrap.producer.model import ProducerValues


class ProducerApp(KafkaApp):
    """
    Producer component

    This producer holds configuration to use as values for the streams bootstrap produce helm chart.
    """

    _type = "producer"
    app: ProducerValues

    class Config(BaseConfig):
        extra = Extra.allow

    @override
    def apply_to_outputs(self, name: str, topic: TopicConfig) -> None:
        match topic.type:
            case OutputTopicTypes.ERROR:
                raise ValueError("Producer apps do not support error topics")
            case _:
                super().apply_to_outputs(name, topic)

    @override
    def set_output_topic(self, topic_name: str) -> None:
        self.app.streams.output_topic = topic_name

    @override
    def add_extra_output_topic(self, topic_name: str, role: str) -> None:
        self.app.streams.extra_output_topics[role] = topic_name

    @override
    def get_helm_chart(self) -> str:
        return f"{self.config.streams_bootstrap_helm_config.repository_name}/{AppType.PRODUCER_APP.value}"

    @property
    @override
    def clean_up_helm_chart(self) -> str:
        return f"{self.config.streams_bootstrap_helm_config.repository_name}/{AppType.CLEANUP_PRODUCER_APP.value}"

    @property
    @override
    def helm_repo_config(self) -> HelmRepoConfig | None:
        return self.config.streams_bootstrap_helm_config

    @override
    def clean(self, dry_run: bool) -> None:
        self._run_clean_up_job(
            values=self.to_helm_values(),
            dry_run=dry_run,
            retain_clean_jobs=self.config.retain_clean_jobs,
        )



from kpops.components.streams_bootstrap.producer.producer_app import ProducerApp
from kpops.components.streams_bootstrap.streams.streams_app import StreamsApp

__all__ = [
    "ProducerApp",
    "StreamsApp",
]



from __future__ import annotations

import json
import logging
import os
from collections.abc import Iterator
from contextlib import suppress
from pathlib import Path

import yaml
from pydantic import BaseModel
from rich.console import Console
from rich.syntax import Syntax

from kpops.cli.pipeline_config import PipelineConfig
from kpops.cli.registry import Registry
from kpops.component_handlers import ComponentHandlers
from kpops.components.base_components.base_defaults_component import update_nested_pair
from kpops.components.base_components.pipeline_component import PipelineComponent
from kpops.utils.yaml_loading import load_yaml_file, substitute

log = logging.getLogger("PipelineGenerator")


class ParsingException(Exception):
    pass


class PipelineComponents(BaseModel):
    components: list[PipelineComponent]


def create_env_components_index(
    environment_components: list[dict],
) -> dict[str, dict]:
    """
    Create an index for all registered components in the project

    :param environment_components: list of components
    :return: component index
    """
    index = {}
    for component in environment_components:
        if "type" not in component or "name" not in component:
            raise ValueError(
                "To override components per environment, every component should at least have a type and a name."
            )
        index[
            PipelineComponent.substitute_component_names(
                component["name"], component["type"], **os.environ
            )
        ] = component
    return index


class Pipeline:
    def __init__(
        self,
        component_list: list[dict],
        environment_components: list[dict],
        registry: Registry,
        config: PipelineConfig,
        handlers: ComponentHandlers,
    ):
        self.components: list[PipelineComponent] = []
        self.handlers = handlers
        self.config = config
        self.registry = registry
        self.parse_components(component_list, environment_components)

        super().__init__()

    @staticmethod
    def pipeline_filename_environment(path: Path, config: PipelineConfig) -> Path:
        """
        Adds the environment name from the PipelineConfig to the pipeline.yaml path
        :param path: Path to pipeline.yaml file
        :param config: The PipelineConfig
        :return: Absolute path to the pipeline_<environment>.yaml
        """
        return path.with_stem(f"{path.stem}_{config.environment}")

    @staticmethod
    def substitute_pipeline_prefix(config: PipelineConfig) -> str:
        """
        Sets the pipline_prefix field in the config with the environment variable pipeline_name.
        :param config: The pipeline config
        :return: The pipeline prefix string
        """
        return substitute(config.pipeline_prefix, dict(os.environ))

    @staticmethod
    def set_pipeline_name_env_vars(base_dir: Path, path: Path) -> None:
        """
        Sets the environment variable pipeline_name relative to the given base_dir.
        Moreover, for each sub-path an environment variable is set.
        For example, for a given path ./data/v1/dev/pipeline.yaml the pipeline_name would be
        set to data-v1-dev. Then the sub environment variables are set:
        pipeline_name_0 = data
        pipeline_name_1 = v1
        pipeline_name_2 = dev
        :param base_dir: Base directory to the pipeline files
        :param path: Path to pipeline.yaml file
        """
        path_without_file = path.resolve().relative_to(base_dir.resolve()).parts[:-1]
        if not path_without_file:
            raise ValueError("The pipeline-base-dir should not equal the pipeline-path")
        pipeline_name = "-".join(path_without_file)
        os.environ["pipeline_name"] = pipeline_name
        for level, parent in enumerate(path_without_file):
            os.environ[f"pipeline_name_{level}"] = parent

    @classmethod
    def load_from_yaml(
        cls,
        base_dir: Path,
        path: Path,
        registry: Registry,
        config: PipelineConfig,
        handlers: ComponentHandlers,
    ) -> Pipeline:
        Pipeline.set_pipeline_name_env_vars(base_dir, path)
        config.pipeline_prefix = Pipeline.substitute_pipeline_prefix(config)

        main_content = load_yaml_file(path, substitution=dict(os.environ))
        if not isinstance(main_content, list):
            raise TypeError(
                f"The pipeline definition {path} should contain a list of components"
            )

        env_content = []
        if (env_file := Pipeline.pipeline_filename_environment(path, config)).exists():
            env_content = load_yaml_file(env_file, substitution=dict(os.environ))
            if not isinstance(env_content, list):
                raise TypeError(
                    f"The pipeline definition {env_file} should contain a list of components"
                )

        pipeline = cls(main_content, env_content, registry, config, handlers)
        return pipeline

    def parse_components(
        self, component_list: list[dict], environment_components: list[dict]
    ) -> None:
        env_components_index = create_env_components_index(environment_components)
        previous_component: PipelineComponent | None = None
        for component in component_list:
            try:
                try:
                    component_type: str = component["type"]
                except KeyError:
                    raise ValueError(
                        "Every component must have a type defined, this component does not have one."
                    )
                component_class = self.registry[component_type]
                inflated_components = self.apply_component(
                    component, component_class, env_components_index, previous_component
                )
                self.populate_pipeline_component_names(inflated_components)
                self.components.extend(inflated_components)
                previous_component = inflated_components.pop()
            except Exception as ex:
                raise ParsingException(
                    f"Error enriching {component['type']} component {component['name']}"
                ) from ex

    def populate_pipeline_component_names(
        self, inflated_components: list[PipelineComponent]
    ) -> None:
        for component in inflated_components:
            component.name = self.config.pipeline_prefix + component.name
            with suppress(
                AttributeError  # Some components like Kafka Connect do not have a name_override attribute
            ):
                if component.app and getattr(component.app, "name_override") is None:
                    setattr(component.app, "name_override", component.name)

    def apply_component(
        self,
        component: dict,
        component_class: type[PipelineComponent],
        env_components_index: dict[str, dict],
        previous_component: PipelineComponent | None,
    ) -> list[PipelineComponent]:
        # Init component for main pipeline
        component_object = self.enrich_component(
            component, component_class, env_components_index
        )

        # weave from topics
        if previous_component and previous_component.to:
            component_object.weave_from_topics(previous_component.to)

        return component_object.inflate()

    def enrich_component(
        self,
        component: dict,
        component_class: type[PipelineComponent],
        env_components_index: dict[str, dict],
    ) -> PipelineComponent:
        component_object: PipelineComponent = component_class(
            handlers=self.handlers, config=self.config, **component
        )
        env_component_definition = env_components_index.get(component_object.name, {})
        pair = update_nested_pair(
            env_component_definition,
            component_object.dict(by_alias=True),
        )

        json_object = self.substitute_component_specific_variables(
            component_object, pair
        )
        return component_class(
            enrich=False,
            handlers=self.handlers,
            config=self.config,
            **json_object,
        )

    @staticmethod
    def substitute_component_specific_variables(
        component_object: PipelineComponent, pair: dict
    ) -> dict:
        # Override component config with component config in pipeline environment definition
        json_object: dict = json.loads(
            substitute(
                json.dumps(pair),
                {
                    "component_type": component_object._type,
                    "component_name": component_object.name,
                },
            )
        )
        return json_object

    def __str__(self) -> str:
        return yaml.dump(
            PipelineComponents(components=self.components).dict(
                exclude_unset=True, exclude_none=True, by_alias=True
            )
        )

    def print_yaml(self, substitution: dict | None = None) -> None:
        syntax = Syntax(
            substitute(str(self), substitution), "yaml", background_color="default"
        )
        Console().print(syntax)

    def save(self, path: Path, substitution: dict | None = None) -> None:
        with open(path, mode="w") as file:
            file.write(substitute(str(self), substitution))

    def __iter__(self) -> Iterator[PipelineComponent]:
        return iter(self.components)



import typer


def yellowify(string: str) -> str:
    return typer.style(string, fg=typer.colors.YELLOW)


def greenify(string: str) -> str:
    return typer.style(string, fg=typer.colors.GREEN)


def magentaify(string: str) -> str:
    return typer.style(string, fg=typer.colors.MAGENTA)



import json
import os
from abc import ABC
from typing import NoReturn

from typing_extensions import override

from kpops.cli.pipeline_config import ENV_PREFIX
from kpops.component_handlers.kafka_connect.connector_handler import KafkaConnectorType
from kpops.component_handlers.kafka_connect.model import KafkaConnectConfig
from kpops.components.base_components.base_defaults_component import deduplicate
from kpops.components.base_components.models.from_section import FromTopic
from kpops.components.base_components.pipeline_component import PipelineComponent


class KafkaConnector(PipelineComponent, ABC):
    _type = "kafka-connect"
    app: KafkaConnectConfig

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.prepare_connector_config()

    def prepare_connector_config(self) -> None:
        """
        Substitute component related variables in config
        """
        substituted_config = self.substitute_component_variables(
            json.dumps(self.app.dict())
        )
        out: dict = json.loads(substituted_config)
        self.app = KafkaConnectConfig(**out)

    @override
    def deploy(self, dry_run: bool) -> None:
        if self.to:
            self.handlers.topic_handler.create_topics(
                to_section=self.to, dry_run=dry_run
            )

            if self.handlers.schema_handler:
                self.handlers.schema_handler.submit_schemas(
                    to_section=self.to, dry_run=dry_run
                )

        self.handlers.connector_handler.create_connector(
            connector_name=self.name, kafka_connect_config=self.app, dry_run=dry_run
        )

    @override
    def destroy(self, dry_run: bool) -> None:
        self.handlers.connector_handler.destroy_connector(
            connector_name=self.name, dry_run=dry_run
        )

    @override
    def clean(self, dry_run: bool) -> None:
        if self.to:
            if self.handlers.schema_handler:
                self.handlers.schema_handler.delete_schemas(
                    to_section=self.to, dry_run=dry_run
                )
            self.handlers.topic_handler.delete_topics(self.to, dry_run=dry_run)


class KafkaSourceConnector(KafkaConnector):
    _type = "kafka-source-connector"

    @override
    def apply_from_inputs(self, name: str, topic: FromTopic) -> NoReturn:
        raise NotImplementedError("Kafka source connector doesn't support FromSection")

    @override
    def reset(self, dry_run: bool) -> None:
        self.__run_kafka_connect_resetter(dry_run)

    @override
    def clean(self, dry_run: bool) -> None:
        super().clean(dry_run)
        self.__run_kafka_connect_resetter(dry_run)

    def __run_kafka_connect_resetter(self, dry_run: bool) -> None:
        self.handlers.connector_handler.clean_connector(
            connector_name=self.name,
            connector_type=KafkaConnectorType.SOURCE,
            dry_run=dry_run,
            retain_clean_jobs=self.config.retain_clean_jobs,
            offset_topic=os.environ[f"{ENV_PREFIX}KAFKA_CONNECT_RESETTER_OFFSET_TOPIC"],
        )


class KafkaSinkConnector(KafkaConnector):
    _type = "kafka-sink-connector"

    @override
    def add_input_topics(self, topics: list[str]) -> None:
        existing_topics: str | None = getattr(self.app, "topics", None)
        topics = existing_topics.split(",") + topics if existing_topics else topics
        topics = deduplicate(topics)
        setattr(self.app, "topics", ",".join(topics))

    @override
    def set_input_pattern(self, name: str) -> None:
        setattr(self.app, "topics.regex", name)

    @override
    def set_error_topic(self, topic_name: str) -> None:
        setattr(self.app, "errors.deadletterqueue.topic.name", topic_name)

    @override
    def reset(self, dry_run: bool) -> None:
        self.__clean_sink_connector(dry_run, delete_consumer_group=False)

    @override
    def clean(self, dry_run: bool) -> None:
        super().clean(dry_run)
        self.__clean_sink_connector(dry_run, delete_consumer_group=True)

    def __clean_sink_connector(
        self, dry_run: bool, delete_consumer_group: bool
    ) -> None:
        self.handlers.connector_handler.clean_connector(
            connector_name=self.name,
            connector_type=KafkaConnectorType.SINK,
            dry_run=dry_run,
            retain_clean_jobs=self.config.retain_clean_jobs,
            delete_consumer_group=delete_consumer_group,
        )



from __future__ import annotations

import logging
import re
from functools import cached_property

from pydantic import BaseModel
from typing_extensions import override

from kpops.component_handlers.helm_wrapper.helm import Helm
from kpops.component_handlers.helm_wrapper.helm_diff import HelmDiff
from kpops.component_handlers.helm_wrapper.model import (
    HelmRepoConfig,
    HelmUpgradeInstallFlags,
)
from kpops.components.base_components.pipeline_component import PipelineComponent
from kpops.utils.colorify import magentaify
from kpops.utils.pydantic import CamelCaseConfig

log = logging.getLogger("KubernetesAppComponent")

KUBERNETES_NAME_CHECK_PATTERN = re.compile(
    r"^(?![0-9]+$)(?!.*-$)(?!-)[a-z0-9-.]{1,253}(?<!_)$"
)


class KubernetesAppConfig(BaseModel):
    namespace: str

    class Config(CamelCaseConfig):
        pass


# TODO: label and annotations
class KubernetesApp(PipelineComponent):
    """Base kubernetes app"""

    _type = "kubernetes-app"
    app: KubernetesAppConfig

    version: str | None = None

    class Config:
        keep_untouched = (cached_property,)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.__check_compatible_name()

    @cached_property
    def helm(self) -> Helm:
        helm = Helm(self.config.helm_config)
        if self.helm_repo_config is not None:
            helm.add_repo(
                self.helm_repo_config.repository_name,
                self.helm_repo_config.url,
                self.helm_repo_config.repo_auth_flags,
            )
        return helm

    @cached_property
    def helm_diff(self) -> HelmDiff:
        return HelmDiff(self.config.helm_diff_config)

    @property
    def helm_release_name(self) -> str:
        """The name for the Helm release. Can be overridden."""
        return self.name

    @property
    def namespace(self) -> str:
        return self.app.namespace

    @property
    def helm_repo_config(self) -> HelmRepoConfig | None:
        return None

    @override
    def deploy(self, dry_run: bool) -> None:
        stdout = self.helm.upgrade_install(
            self.helm_release_name,
            self.get_helm_chart(),
            dry_run,
            self.namespace,
            self.to_helm_values(),
            HelmUpgradeInstallFlags(version=self.version),
        )
        if dry_run and self.helm_diff.config.enable:
            self.print_helm_diff(stdout)

    @override
    def destroy(self, dry_run: bool) -> None:
        stdout = self.helm.uninstall(
            self.namespace,
            self.helm_release_name,
            dry_run,
        )

        if stdout:
            log.info(magentaify(stdout))

    def to_helm_values(self) -> dict:
        return self.app.dict(by_alias=True, exclude_none=True, exclude_unset=True)

    def print_helm_diff(self, stdout: str) -> None:
        current_release = self.helm.get_manifest(self.helm_release_name, self.namespace)
        new_release = Helm.load_helm_manifest(stdout)
        helm_diff = HelmDiff.get_diff(current_release, new_release)
        self.helm_diff.log_helm_diff(helm_diff, log)

    def get_helm_chart(self) -> str:
        raise NotImplementedError(
            f"Please implement the get_helm_chart() method of the {self.__module__} module."
        )

    def __check_compatible_name(self) -> None:
        if not bool(KUBERNETES_NAME_CHECK_PATTERN.match(self.name)):  # TODO: SMARTER
            raise ValueError(
                f"The component name {self.name} is invalid for Kubernetes."
            )



from collections.abc import Mapping
from pathlib import Path
from string import Template
from typing import Any

import yaml
from cachetools import cached
from cachetools.keys import hashkey


def generate_hashkey(file_path: Path, substitution: dict | None = None) -> tuple:
    if substitution is None:
        substitution = {}
    return hashkey(str(file_path) + str(sorted(substitution.items())))


@cached(cache={}, key=generate_hashkey)
def load_yaml_file(
    file_path: Path, *, substitution: dict | None = None
) -> dict | list[dict]:
    with open(file_path) as yaml_file:
        return yaml.load(substitute(yaml_file.read(), substitution), Loader=yaml.Loader)


def substitute(input: str, substitution: Mapping[str, Any] | None = None) -> str:
    """
    Substitute $-placeholders in input using template string.
    :param input: The raw input containing $-placeholders
    :param substitution: The key-value mapping containing substitutions
    :return: Substituted input string
    """
    if not substitution:
        return input
    return Template(input).safe_substitute(**substitution)



import logging
import signal
from typing import Callable, NoReturn, TypeVar

log = logging.getLogger("Timeout")

T = TypeVar("T")


def timeout(func: Callable[..., T], *, secs: int = 0) -> T | None:
    """
    Sets a timeout for a given lambda function
    :param func: The callable function
    :param secs: The timeout in seconds
    """

    def handler(signum, frame) -> NoReturn:
        raise TimeoutError()

    # set the timeout handler
    signal.signal(signal.SIGALRM, handler)
    signal.alarm(secs)
    try:
        return func()
    except TimeoutError:
        log.error(
            f"Kafka Connect operation {func.__name__} timed out after {secs} seconds. To increase the duration, set the `timeout` option in config.yaml."
        )
    finally:
        signal.alarm(0)



from __future__ import annotations

import logging
import re
import subprocess
import tempfile
from collections.abc import Iterator
from typing import Iterable

import yaml

from kpops.component_handlers.helm_wrapper.exception import ReleaseNotFoundException
from kpops.component_handlers.helm_wrapper.model import (
    HelmConfig,
    HelmTemplate,
    HelmUpgradeInstallFlags,
    RepoAuthFlags,
    YamlReader,
)

log = logging.getLogger("Helm")


class Helm:
    def __init__(self, helm_config: HelmConfig) -> None:
        self._context = helm_config.context
        self._debug = helm_config.debug

    def add_repo(
        self,
        repository_name: str,
        repository_url: str,
        repo_auth_flags: RepoAuthFlags = RepoAuthFlags(),
    ) -> None:
        command = [
            "helm",
            "repo",
            "add",
            repository_name,
            repository_url,
        ]

        if repo_auth_flags.username and repo_auth_flags.password:
            command.extend(
                [
                    "--username",
                    repo_auth_flags.username,
                    "--password",
                    repo_auth_flags.password,
                ]
            )

        command = Helm.__extend_tls_config(command, repo_auth_flags)

        try:
            self.__execute(command)
        except Exception as e:
            if (
                len(e.args) == 1
                and re.match(
                    "Error: repository name (.*) already exists, please specify a different name",
                    e.args[0],
                )
                is not None
            ):
                log.error(f"Could not add repository {repository_name}. {e}")
            else:
                raise e

        self.__execute(["helm", "repo", "update"])

    def upgrade_install(
        self,
        release_name: str,
        chart: str,
        dry_run: bool,
        namespace: str,
        values: dict,
        flags: HelmUpgradeInstallFlags = HelmUpgradeInstallFlags(),
    ) -> str:
        """
        Prepares and executes the helm upgrade install command
        """
        with tempfile.NamedTemporaryFile("w") as values_file:
            yaml.safe_dump(values, values_file)

            command = ["helm"]
            command.extend(
                [
                    "upgrade",
                    release_name,
                    chart,
                    "--install",
                    f"--timeout={flags.timeout}",
                    "--namespace",
                    namespace,
                    "--values",
                    values_file.name,
                ]
            )
            command = Helm.__extend_tls_config(command, flags.repo_auth_flags)

            command = Helm.__enrich_upgrade_install_command(command, dry_run, flags)
            return self.__execute(command)

    def uninstall(
        self,
        namespace: str,
        release_name: str,
        dry_run: bool,
    ) -> str | None:
        """
        Prepares and executes the helm uninstall command
        """
        command = [
            "helm",
            "uninstall",
            release_name,
            "--namespace",
            namespace,
        ]
        if dry_run:
            command.append("--dry-run")
        try:
            return self.__execute(command)
        except ReleaseNotFoundException:
            log.warning(
                f"Release with name {release_name} not found. Could not uninstall app."
            )

    def get_manifest(self, release_name: str, namespace: str) -> Iterable[HelmTemplate]:
        command = [
            "helm",
            "get",
            "manifest",
            release_name,
            "--namespace",
            namespace,
        ]

        try:
            stdout = self.__execute(command=command)
            return Helm.load_helm_manifest(stdout)
        except ReleaseNotFoundException:
            return ()

    @staticmethod
    def load_helm_manifest(yaml_contents: str) -> Iterator[HelmTemplate]:
        is_beginning: bool = False
        template_name = None
        current_yaml_doc: list[str] = []
        for line in YamlReader(yaml_contents):
            if line.startswith("---"):
                is_beginning = True
                if template_name and current_yaml_doc:
                    yield HelmTemplate.load(template_name, "\n".join(current_yaml_doc))
                    template_name = None
                    current_yaml_doc.clear()
            elif is_beginning:
                template_name = HelmTemplate.parse_source(line)
                is_beginning = False
            else:
                current_yaml_doc.append(line)

    @staticmethod
    def __extend_tls_config(
        command: list[str], repo_auth_flags: RepoAuthFlags
    ) -> list[str]:
        if repo_auth_flags.ca_file:
            command.extend(["--ca-file", str(repo_auth_flags.ca_file)])
        if repo_auth_flags.insecure_skip_tls_verify:
            command.append("--insecure-skip-tls-verify")
        return command

    @staticmethod
    def __enrich_upgrade_install_command(
        command: list[str],
        dry_run: bool,
        helm_command_config: HelmUpgradeInstallFlags,
    ) -> list[str]:
        if dry_run:
            command.append("--dry-run")
        if helm_command_config.force:
            command.append("--force")
        if helm_command_config.wait:
            command.append("--wait")
        if helm_command_config.wait_for_jobs:
            command.append("--wait-for-jobs")
        if helm_command_config.version:
            command.extend(["--version", helm_command_config.version])
        return command

    def __execute(self, command: list[str]) -> str:
        command = self.__set_global_flags(command)
        log.debug(f"Executing {' '.join(command)}")
        process = subprocess.run(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        Helm.parse_helm_command_stderr_output(process.stderr)
        log.debug(process.stdout)
        return process.stdout

    def __set_global_flags(self, command: list[str]) -> list[str]:
        if self._context:
            log.debug(f"Changing the Kubernetes context to {self._context}")
            command.extend(["--kube-context", self._context])
        if self._debug:
            log.debug("Enabling verbose mode.")
            command.append("--debug")
        return command

    @staticmethod
    def parse_helm_command_stderr_output(stderr: str) -> None:
        for line in stderr.splitlines():
            lower = line.lower()
            if "release: not found" in lower:
                raise ReleaseNotFoundException()
            elif "error" in lower:
                raise RuntimeError(stderr)
            elif "warning" in lower:
                log.warning(line)



from difflib import Differ
from enum import Enum
from typing import Any, Iterable, Iterator, Sequence

import typer
import yaml
from dictdiffer import diff, patch
from pydantic import BaseModel

differ = Differ()


class DiffType(str, Enum):
    ADD = "add"
    CHANGE = "change"
    REMOVE = "remove"


class Change(BaseModel):
    old_value: Any | None
    new_value: Any | None


class Diff(BaseModel):
    diff_type: DiffType
    key: str
    change: Change


def render_diff(d1: dict, d2: dict, ignore: set[str] | None = None) -> str | None:
    differences = list(diff(d1, d2, ignore=ignore))
    if not differences:
        return None

    d2_filtered: dict = patch(differences, d1)
    return "".join(
        colorize_diff(
            differ.compare(
                to_yaml(d1) if d1 else "",
                to_yaml(d2_filtered) if d2_filtered else "",
            )
        )
    )


def get_diff(d1: dict, d2: dict, ignore: set[str] | None = None) -> list[Diff]:
    differences = list(diff(d1, d2, ignore=ignore))
    diff_list = []
    for difference in differences:
        if difference[0] == DiffType.ADD.value:
            for key, change in difference[2]:
                diff_object = Diff(
                    diff_type=DiffType[difference[0].upper()],
                    key=__get_key(difference[1], key),
                    change=Change(old_value=None, new_value=change),
                )
                diff_list.append(diff_object)

        elif difference[0] == DiffType.REMOVE.value:
            for key, change in difference[2]:
                diff_object = Diff(
                    diff_type=DiffType[difference[0].upper()],
                    key=__get_key(difference[1], key),
                    change=Change(old_value=change, new_value=None),
                )
                diff_list.append(diff_object)

        elif difference[0] == DiffType.CHANGE.value:
            diff_object = Diff(
                diff_type=DiffType[difference[0].upper()],
                key=__get_key(difference[1], ""),
                change=Change(old_value=difference[2][0], new_value=difference[2][1]),
            )
            diff_list.append(diff_object)
        else:
            raise ValueError(
                f"Unsupported diff type of {difference[0]}. Suppoerted diff types are: ADD, CHANGE, REMOVE."
            )

    return diff_list


def __get_key(key_1: list[Any] | str, key_2: str) -> str:
    """
    Generates a string that points to the changed key in the dictionary.
    """
    if len(key_1) > 1 and type(key_1) == list:
        return f"{key_1[0]}[{key_1[1]}]"
    else:
        if key_1 == "":
            return key_2
        elif key_2 == "":
            return "".join(key_1)
        return f"{key_1}.{key_2}"


def colorize_diff(input: Iterable[str]) -> Iterator[str]:
    for line in input:
        yield colorize_line(line)


def colorize_line(line: str) -> str:
    if line.startswith("-"):
        return typer.style(line, fg=typer.colors.RED)
    if line.startswith("+"):
        return typer.style(line, fg=typer.colors.GREEN)
    if line.startswith("?"):
        return typer.style(line, fg=typer.colors.YELLOW)
    return line


def to_yaml(data: dict) -> Sequence[str]:
    return yaml.dump(data, sort_keys=True).splitlines(keepends=True)



from dataclasses import dataclass, field
from pathlib import Path
from typing import Iterator

import yaml


@dataclass
class HelmDiffConfig:
    enable: bool = False
    ignore: set[str] = field(
        default_factory=set,
    )


@dataclass
class RepoAuthFlags:
    username: str | None = None
    password: str | None = None
    ca_file: Path | None = None
    insecure_skip_tls_verify: bool = False


@dataclass
class HelmRepoConfig:
    repository_name: str
    url: str
    repo_auth_flags: RepoAuthFlags = field(default_factory=RepoAuthFlags)


@dataclass
class HelmConfig:
    context: str | None = None
    debug: bool = False


@dataclass
class HelmUpgradeInstallFlags:
    force: bool = False
    repo_auth_flags: RepoAuthFlags = field(default_factory=RepoAuthFlags)
    timeout: str = "5m0s"
    version: str | None = None
    wait: bool = True
    wait_for_jobs: bool = False


HELM_SOURCE_PREFIX = "# Source: "


@dataclass
class HelmTemplate:
    filepath: str
    template: dict

    @staticmethod
    def parse_source(source: str) -> str:
        """Parse source path from comment at the beginning of the YAML doc.
        Example: # Source: chart/templates/serviceaccount.yaml
        """
        if not source.startswith(HELM_SOURCE_PREFIX):
            raise ValueError("Not a valid Helm template source")
        return source.removeprefix(HELM_SOURCE_PREFIX).strip()

    @classmethod
    def load(cls, filepath: str, content: str):
        template = yaml.load(content, yaml.Loader)
        return cls(filepath, template)


@dataclass
class YamlReader:
    content: str

    def __iter__(self) -> Iterator[str]:
        # discard all output before template documents
        index = self.content.index("---")
        self.content = self.content[index:-1]
        yield from self.content.splitlines()
        yield "---"  # add final divider to make parsing easier



from pydantic import BaseConfig, Extra

from kpops.components.base_components.kafka_app import (
    KafkaAppConfig,
    KafkaStreamsConfig,
)


class ProducerStreamsConfig(KafkaStreamsConfig):
    extra_output_topics: dict[str, str] = {}
    output_topic: str | None


class ProducerValues(KafkaAppConfig):
    streams: ProducerStreamsConfig

    class Config(BaseConfig):
        extra = Extra.allow



import logging
import time
from time import sleep

import requests

from kpops.component_handlers.kafka_connect.exception import (
    ConnectorNotFoundException,
    KafkaConnectError,
)
from kpops.component_handlers.kafka_connect.model import (
    KafkaConnectConfig,
    KafkaConnectConfigErrorResponse,
    KafkaConnectResponse,
)

HEADERS = {"Accept": "application/json", "Content-Type": "application/json"}

log = logging.getLogger("KafkaConnectAPI")


class ConnectWrapper:
    """
    Wraps Kafka Connect APIs
    """

    def __init__(self, host: str | None):
        if not host:
            error_message = (
                "The Kafka Connect host is not set. Please set the host in the config."
            )
            log.error(error_message)
            raise RuntimeError(error_message)
        self._host: str = host

    @property
    def host(self) -> str:
        return self._host

    def create_connector(
        self, connector_name: str, kafka_connect_config: KafkaConnectConfig
    ) -> KafkaConnectResponse:
        """
        Creates a new connector
        API Reference: https://docs.confluent.io/platform/current/connect/references/restapi.html#post--connectors
        :param connector_name: The name of the connector
        :param kafka_connect_config: The config of the connector
        :return: The current connector info if successful
        """
        config_json = kafka_connect_config.dict(exclude_none=True, exclude_unset=True)
        connect_data = {"name": connector_name, "config": config_json}
        response = requests.post(
            url=f"{self._host}/connectors", headers=HEADERS, json=connect_data
        )
        if response.status_code == requests.status_codes.codes.created:
            log.info(f"Connector {connector_name} created.")
            log.debug(response.json())
            return KafkaConnectResponse(**response.json())
        elif response.status_code == requests.status_codes.codes.conflict:
            log.warning(
                "Rebalancing in progress while creating a connector... Retrying..."
            )
            time.sleep(1)
            self.create_connector(connector_name, kafka_connect_config)
        raise KafkaConnectError(response)

    def get_connector(self, connector_name: str) -> KafkaConnectResponse:
        """
        Get information about the connector.
        API Reference: https://docs.confluent.io/platform/current/connect/references/restapi.html#get--connectors-(string-name)
        :param connector_name: Nameof the crated connector
        :return: Information about the connector
        """
        response = requests.get(
            url=f"{self._host}/connectors/{connector_name}", headers=HEADERS
        )
        if response.status_code == requests.status_codes.codes.ok:
            log.info(f"Connector {connector_name} exists.")
            log.debug(response.json())
            return KafkaConnectResponse(**response.json())
        elif response.status_code == requests.status_codes.codes.not_found:
            log.info(f"The named connector {connector_name} does not exists.")
            raise ConnectorNotFoundException()
        elif response.status_code == requests.status_codes.codes.conflict:
            log.warning(
                "Rebalancing in progress while getting a connector... Retrying..."
            )
            sleep(1)
            self.get_connector(connector_name)
        raise KafkaConnectError(response)

    def update_connector_config(
        self, connector_name: str, kafka_connect_config: KafkaConnectConfig
    ) -> KafkaConnectResponse:
        """
        Create a new connector using the given configuration, or update the configuration for an existing connector.
        :param connector_name: Name of the created connector
        :param kafka_connect_config: Configuration parameters for the connector.
        :return: Information about the connector after the change has been made.
        """
        config_json = kafka_connect_config.dict(exclude_none=True, exclude_unset=True)
        response = requests.put(
            url=f"{self._host}/connectors/{connector_name}/config",
            headers=HEADERS,
            json=config_json,
        )
        data: dict = response.json()
        if response.status_code == requests.status_codes.codes.ok:
            log.info(f"Config for connector {connector_name} updated.")
            log.debug(data)
            return KafkaConnectResponse(**data)
        if response.status_code == requests.status_codes.codes.created:
            log.info(f"Connector {connector_name} created.")
            log.debug(data)
            return KafkaConnectResponse(**data)
        elif response.status_code == requests.status_codes.codes.conflict:
            log.warning(
                "Rebalancing in progress while updating a connector... Retrying..."
            )
            sleep(1)
            self.update_connector_config(connector_name, kafka_connect_config)
        raise KafkaConnectError(response)

    def validate_connector_config(
        self, kafka_connect_config: KafkaConnectConfig
    ) -> list[str]:
        """
        Validate connector config using the given configuration
        :param connector_name: Name of the created connector
        :param kafka_connect_config: Configuration parameters for the connector.
        :return:
        """
        config_json = kafka_connect_config.dict(exclude_none=True, exclude_unset=True)
        connector_class = ConnectWrapper.get_connector_class_name(config_json)

        response = requests.put(
            url=f"{self._host}/connector-plugins/{connector_class}/config/validate",
            headers=HEADERS,
            json=config_json,
        )

        if response.status_code == requests.status_codes.codes.ok:
            kafka_connect_error_response = KafkaConnectConfigErrorResponse(
                **response.json()
            )

            errors = []
            if kafka_connect_error_response.error_count > 0:
                for config in kafka_connect_error_response.configs:
                    if len(config.value.errors) > 0:
                        for error in config.value.errors:
                            errors.append(
                                f"Found error for field {config.value.name}: {error}"
                            )
            return errors
        raise KafkaConnectError(response)

    def delete_connector(self, connector_name: str) -> None:
        """
        Deletes a connector, halting all tasks and deleting its configuration.
        API Reference:https://docs.confluent.io/platform/current/connect/references/restapi.html#delete--connectors-(string-name)-
        """
        response = requests.delete(
            url=f"{self._host}/connectors/{connector_name}", headers=HEADERS
        )
        if response.status_code == requests.status_codes.codes.no_content:
            log.info(f"Connector {connector_name} deleted.")
            return
        elif response.status_code == requests.status_codes.codes.not_found:
            log.info(f"The named connector {connector_name} does not exists.")
            raise ConnectorNotFoundException()
        elif response.status_code == requests.status_codes.codes.conflict:
            log.warning(
                "Rebalancing in progress while deleting a connector... Retrying..."
            )
            sleep(1)
            self.delete_connector(connector_name)
        raise KafkaConnectError(response)

    @staticmethod
    def get_connector_class_name(config_json: dict[str, str]) -> str:
        task_class = config_json["connector.class"]
        return task_class.split(".")[-1]



from __future__ import annotations

from typing import TYPE_CHECKING

from kpops.component_handlers.kafka_connect.connector_handler import ConnectorHandler

if TYPE_CHECKING:
    from kpops.component_handlers.schema_handler.schema_handler import SchemaHandler
    from kpops.component_handlers.topic.handler import TopicHandler


class ComponentHandlers:
    def __init__(
        self,
        schema_handler: SchemaHandler | None,
        connector_handler: ConnectorHandler,
        topic_handler: TopicHandler,
    ) -> None:
        self.schema_handler = schema_handler
        self.connector_handler = connector_handler
        self.topic_handler = topic_handler



from kpops.components.base_components.kafka_app import KafkaApp
from kpops.components.base_components.kafka_connect import (
    KafkaSinkConnector,
    KafkaSourceConnector,
)
from kpops.components.base_components.kubernetes_app import KubernetesApp
from kpops.components.base_components.pipeline_component import PipelineComponent

__all__ = (
    "KafkaApp",
    "KafkaSinkConnector",
    "KafkaSourceConnector",
    "KubernetesApp",
    "PipelineComponent",
)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, TypeAlias

from schema_registry.client.schema import AvroSchema, JsonSchema

Schema: TypeAlias = AvroSchema | JsonSchema


class SchemaProvider(ABC):
    @abstractmethod
    def provide_schema(self, schema_class: str, models: dict[str, Any]) -> Schema:
        ...



from enum import Enum
from typing import Any

from pydantic import BaseConfig, BaseModel, Extra, Field, root_validator


class OutputTopicTypes(str, Enum):
    """
    Types of output topic
    error (error topic), output (output topic), and extra topics. Every extra topic must have a role.
    """

    ERROR = "error"
    OUTPUT = "output"
    EXTRA = "extra"


class TopicConfig(BaseModel):
    """Configures a topic"""

    type: OutputTopicTypes = Field(...)
    key_schema: str | None = Field(default=None, alias="keySchema")
    value_schema: str | None = Field(default=None, alias="valueSchema")
    partitions_count: int | None = None
    replication_factor: int | None = None
    configs: dict[str, str] = {}
    role: str | None = None

    class Config(BaseConfig):
        extra = Extra.forbid
        allow_population_by_field_name = True
        use_enum_values = True

    @root_validator
    def extra_topic_role(cls, values):
        is_extra_topic: bool = values["type"] == OutputTopicTypes.EXTRA
        if is_extra_topic and not values.get("role"):
            raise ValueError(
                "If you define an extra output topic, you have to define a role."
            )
        if not is_extra_topic and values.get("role"):
            raise ValueError(
                "If you do not define a output topic, the role is unnecessary. (This topic is either an output topic "
                "without a role or an error topic)"
            )
        return values


class ToSection(BaseModel):
    models: dict[
        str, Any
    ] = (
        {}
    )  # any because snapshot versions must be supported  # TODO: really multiple models?
    topics: dict[str, TopicConfig]

    class Config(BaseConfig):
        extra = Extra.allow



import humps
from pydantic import BaseConfig


def to_camel(field: str) -> str:
    return humps.camelize(field)  # type: ignore


class CamelCaseConfig(BaseConfig):
    alias_generator = to_camel
    allow_population_by_field_name = True



from __future__ import annotations

from pydantic import BaseConfig, Extra
from typing_extensions import override

from kpops.component_handlers.helm_wrapper.model import HelmRepoConfig
from kpops.components.base_components.kafka_app import KafkaApp
from kpops.components.streams_bootstrap.app_type import AppType
from kpops.components.streams_bootstrap.streams.model import StreamsAppConfig


class StreamsApp(KafkaApp):
    """
    StreamsApp component that configures a streams bootstrap app
    """

    _type = "streams-app"
    app: StreamsAppConfig

    class Config(BaseConfig):
        extra = Extra.allow

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.__substitute_autoscaling_topic_names()

    @override
    def add_input_topics(self, topics: list[str]) -> None:
        self.app.streams.add_input_topics(topics)

    @override
    def add_extra_input_topic(self, role: str, topics: list[str]) -> None:
        self.app.streams.add_extra_input_topics(role, topics)

    @override
    def set_input_pattern(self, name: str) -> None:
        self.app.streams.input_pattern = name

    @override
    def add_extra_input_pattern(self, role: str, topic: str) -> None:
        self.app.streams.extra_input_patterns[role] = topic

    @override
    def set_output_topic(self, topic_name: str) -> None:
        self.app.streams.output_topic = topic_name

    @override
    def set_error_topic(self, topic_name: str) -> None:
        self.app.streams.error_topic = topic_name

    @override
    def add_extra_output_topic(self, topic_name: str, role: str) -> None:
        self.app.streams.extra_output_topics[role] = topic_name

    @override
    def get_helm_chart(self) -> str:
        return f"{self.config.streams_bootstrap_helm_config.repository_name}/{AppType.STREAMS_APP.value}"

    @property
    @override
    def helm_repo_config(self) -> HelmRepoConfig | None:
        return self.config.streams_bootstrap_helm_config

    @property
    @override
    def clean_up_helm_chart(self) -> str:
        return f"{self.config.streams_bootstrap_helm_config.repository_name}/{AppType.CLEANUP_STREAMS_APP.value}"

    @override
    def reset(self, dry_run: bool) -> None:
        self.__run_streams_clean_up_job(dry_run, delete_output=False)

    @override
    def clean(self, dry_run: bool) -> None:
        self.__run_streams_clean_up_job(dry_run, delete_output=True)

    def __run_streams_clean_up_job(self, dry_run: bool, delete_output: bool) -> None:
        values = self.to_helm_values()
        values["streams"]["deleteOutput"] = delete_output
        self._run_clean_up_job(
            values=values,
            dry_run=dry_run,
            retain_clean_jobs=self.config.retain_clean_jobs,
        )

    def __substitute_autoscaling_topic_names(self) -> None:
        if not self.app.autoscaling:
            return
        self.app.autoscaling.topics = [
            self.substitute_component_variables(topic)
            for topic in self.app.autoscaling.topics
        ]
        self.app.autoscaling.consumergroup = self.substitute_component_variables(
            self.app.autoscaling.consumergroup
        )



import logging

import requests

log = logging.getLogger("RequestException")


class RequestsException(Exception):
    def __init__(self, response: requests.Response) -> None:
        self.error_code = response.status_code
        self.error_msg = "Something went wrong!"
        try:
            log.error(
                f"The request responded with the code {self.error_code}. Error body: {response.json()}"
            )
            response.raise_for_status()
        except requests.HTTPError as e:
            self.error_msg = str(e)
            log.error(f"More information: {self.error_msg}")
        super().__init__()



from __future__ import annotations

import logging

from pydantic import BaseModel, Extra
from typing_extensions import override

from kpops.component_handlers.helm_wrapper.helm import Helm
from kpops.component_handlers.helm_wrapper.model import HelmUpgradeInstallFlags
from kpops.component_handlers.helm_wrapper.utils import trim_release_name
from kpops.components.base_components.kubernetes_app import (
    KubernetesApp,
    KubernetesAppConfig,
)
from kpops.utils.pydantic import CamelCaseConfig
from kpops.utils.yaml_loading import substitute

log = logging.getLogger("KafkaApp")


class KafkaStreamsConfig(BaseModel):
    brokers: str
    schema_registry_url: str | None = None

    class Config(CamelCaseConfig):
        extra = Extra.allow


class KafkaAppConfig(KubernetesAppConfig):
    streams: KafkaStreamsConfig
    name_override: str | None


class KafkaApp(KubernetesApp):
    """
    Base component for kafka-based components.
    Producer or streaming apps should inherit from this class.
    """

    _type = "kafka-app"
    app: KafkaAppConfig

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.app.streams.brokers = substitute(
            self.app.streams.brokers, {"broker": self.config.broker}
        )
        if self.app.streams.schema_registry_url:
            self.app.streams.schema_registry_url = substitute(
                self.app.streams.schema_registry_url,
                {"schema_registry_url": self.config.schema_registry_url},
            )

    @property
    def clean_up_helm_chart(self) -> str:
        raise NotImplementedError()

    @override
    def deploy(self, dry_run: bool) -> None:
        if self.to:
            self.handlers.topic_handler.create_topics(
                to_section=self.to, dry_run=dry_run
            )

            if self.handlers.schema_handler:
                self.handlers.schema_handler.submit_schemas(
                    to_section=self.to, dry_run=dry_run
                )
        super().deploy(dry_run)

    def _run_clean_up_job(
        self,
        values: dict,
        dry_run: bool,
        retain_clean_jobs: bool = False,
    ) -> None:
        """
        Cleans an app using the respective cleanup job
        :param values: The value YAML for the chart
        :param dry_run: Dry run command
        :param retain_clean_jobs: Whether to retain the cleanup job
        :return:
        """
        suffix = "-clean"
        clean_up_release_name = trim_release_name(
            self.helm_release_name + suffix, suffix
        )
        log.info(f"Uninstall old cleanup job for {clean_up_release_name}")

        self.__uninstall_clean_up_job(clean_up_release_name, dry_run)

        log.info(f"Init cleanup job for {clean_up_release_name}")

        stdout = self.__install_clean_up_job(
            clean_up_release_name, suffix, values, dry_run
        )

        if dry_run and self.helm_diff.config.enable:
            current_release = self.helm.get_manifest(
                clean_up_release_name, self.namespace
            )
            new_release = Helm.load_helm_manifest(stdout)
            self.helm_diff.get_diff(current_release, new_release)

        if not retain_clean_jobs:
            log.info(f"Uninstall cleanup job for {clean_up_release_name}")
            self.__uninstall_clean_up_job(clean_up_release_name, dry_run)

    def __uninstall_clean_up_job(self, release_name: str, dry_run: bool) -> None:
        self.helm.uninstall(self.namespace, release_name, dry_run)

    def __install_clean_up_job(
        self,
        release_name: str,
        suffix: str,
        values: dict,
        dry_run: bool,
    ) -> str:
        clean_up_release_name = trim_release_name(release_name, suffix)
        return self.helm.upgrade_install(
            clean_up_release_name,
            self.clean_up_helm_chart,
            dry_run,
            self.namespace,
            values,
            HelmUpgradeInstallFlags(
                version=self.version,
                wait=True,
                wait_for_jobs=True,
            ),
        )



from kpops.components.base_components import (
    KafkaApp,
    KafkaSinkConnector,
    KafkaSourceConnector,
    KubernetesApp,
)
from kpops.components.streams_bootstrap import ProducerApp, StreamsApp

__all__ = (
    "KafkaApp",
    "KafkaSinkConnector",
    "KafkaSourceConnector",
    "KubernetesApp",
    "ProducerApp",
    "StreamsApp",
)



from pathlib import Path

from pydantic import BaseConfig, BaseSettings, Field
from pydantic.env_settings import SettingsSourceCallable

from kpops.component_handlers.helm_wrapper.model import (
    HelmConfig,
    HelmDiffConfig,
    HelmRepoConfig,
)
from kpops.utils.yaml_loading import load_yaml_file

ENV_PREFIX = "KPOPS_"


class TopicNameConfig(BaseSettings):
    default_output_topic_name: str = Field(
        default="${pipeline_name}-${component_name}",
        description="Configures the value for the variable ${output_topic_name}",
    )
    default_error_topic_name: str = Field(
        default="${pipeline_name}-${component_name}-error",
        description="Configures the value for the variable ${error_topic_name}",
    )


class KafkaConnectResetterHelmConfig(BaseSettings):
    helm_config: HelmRepoConfig = Field(
        default=HelmRepoConfig(
            repository_name="bakdata-kafka-connect-resetter",
            url="https://bakdata.github.io/kafka-connect-resetter/",
        ),
        description="Configuration of Kafka connect resetter Helm Chart",
    )
    version: str = "1.0.4"
    helm_values: dict = Field(
        default={},
        description="Overriding Kafka Connect Resetter Helm values. E.g. to override the Image Tag etc.",
    )
    namespace: str = Field(default="")


class PipelineConfig(BaseSettings):
    defaults_path: Path = Field(
        default=...,
        description="The path to the folder containing the defaults file and the environment defaults files.",
    )
    environment: str = Field(
        default=...,
        env=f"{ENV_PREFIX}ENVIRONMENT",
        example="development",
        description="The environment you want to generate and deploy the pipeline to. "
        "Suffix your environment files with this value (e.g. defaults_development.yaml for environment=development).",
    )
    broker: str = Field(
        default=...,
        env=f"{ENV_PREFIX}KAFKA_BROKER",
        description="The kafka broker address.",
    )
    defaults_filename_prefix: str = Field(
        default="defaults",
        description="The name of the defaults file and the prefix of the defaults environment file.",
    )
    topic_name_config: TopicNameConfig = Field(
        default=TopicNameConfig(),
        description="Configure the topic name variables you can use in the pipeline definition.",
    )
    kafka_rest_host: str | None = Field(
        default=None,
        env=f"{ENV_PREFIX}REST_PROXY_HOST",
        example="http://localhost:8082",
        description="Address to the rest proxy REST API.",
    )
    kafka_connect_host: str | None = Field(
        default=None,
        env=f"{ENV_PREFIX}CONNECT_HOST",
        example="http://localhost:8083",
        description="Address to the kafka connect REST API.",
    )
    timeout: int = Field(
        default=300,
        env=f"{ENV_PREFIX}TIMEOUT",
        description="The timeout in seconds that specifies when actions like deletion or deploy timeout.",
    )
    pipeline_prefix: str = Field(
        default="${pipeline_name}-",
        env=f"{ENV_PREFIX}PIPELINE_PREFIX",
        description="Pipeline prefix that will prefix every component name. If you wish to not have any prefix you can specify an empty string.",
    )

    helm_config: HelmConfig = Field(default=HelmConfig())
    helm_diff_config: HelmDiffConfig = Field(default=HelmDiffConfig())

    streams_bootstrap_helm_config: HelmRepoConfig = Field(
        default=HelmRepoConfig(
            repository_name="bakdata-streams-bootstrap",
            url="https://bakdata.github.io/streams-bootstrap/",
        ),
        description="Configuration for Streams Bootstrap Helm Charts",
    )
    kafka_connect_resetter_config: KafkaConnectResetterHelmConfig = Field(
        default=KafkaConnectResetterHelmConfig(),
        description="Configuration of kafka connect resetter helm chart and values. "
        "This is used for cleaning/resettting Kafka connectors, see https://github.com/bakdata/kafka-connect-resetter",
    )
    retain_clean_jobs: bool = Field(
        default=False,
        env=f"{ENV_PREFIX}RETAIN_CLEAN_JOBS",
        description="Whether to retain clean up jobs in the cluster or uninstall the, after completion.",
    )
    schema_registry_url: str | None = Field(
        default=None,
        example="http://localhost:8081",
        env=f"{ENV_PREFIX}SCHEMA_REGISTRY_URL",
        description="The URL to schema registry.",
    )

    class Config(BaseConfig):
        config_path: Path = Path("config.yaml")
        env_file = ".env"
        env_file_encoding = "utf-8"

        @classmethod
        def customise_sources(
            cls,
            init_settings: SettingsSourceCallable,
            env_settings: SettingsSourceCallable,
            file_secret_settings: SettingsSourceCallable,
        ):
            return (
                init_settings,
                yaml_config_settings_source,
                env_settings,
                file_secret_settings,
            )


def yaml_config_settings_source(settings: PipelineConfig) -> dict | list:
    path_to_config = settings.Config.config_path
    if path_to_config.exists():
        return load_yaml_file(path_to_config)
    return {}



from __future__ import annotations

import os

from pydantic import BaseConfig, Extra, Field

from kpops.components.base_components.base_defaults_component import (
    BaseDefaultsComponent,
)
from kpops.components.base_components.models.from_section import (
    FromSection,
    FromTopic,
    InputTopicTypes,
)
from kpops.components.base_components.models.to_section import (
    OutputTopicTypes,
    TopicConfig,
    ToSection,
)
from kpops.utils.yaml_loading import substitute


class PipelineComponent(BaseDefaultsComponent):
    name: str
    from_: FromSection | None = Field(default=None, alias="from")
    app: object | None = None
    to: ToSection | None = None

    class Config(BaseConfig):
        extra = Extra.allow

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.substitute_output_topic_names()
        self.substitute_name()
        self.set_input_topics()
        self.set_output_topics()

    def substitute_output_topic_names(self):
        """
        Substitutes component and topic sepcific names in output topics
        """
        if self.to:
            updated_to = {}
            for name, topic in self.to.topics.items():
                name = self.substitute_component_variables(name)
                updated_to[name] = topic
            self.to.topics = updated_to

    @staticmethod
    def substitute_component_names(key: str, _type: str, **kwargs) -> str:
        return substitute(key, {"component_type": _type, **kwargs})

    def substitute_component_variables(self, topic_name: str) -> str:
        """
        Substitutes component, env and topic specific variables in the topic name

        :param topic_name: topic name
        :return: final topic name
        """
        error_topic_name = self.substitute_component_names(
            self.config.topic_name_config.default_error_topic_name,
            self._type,
            **os.environ,
        )
        output_topic_name = self.substitute_component_names(
            self.config.topic_name_config.default_output_topic_name,
            self._type,
            **os.environ,
        )
        return self.substitute_component_names(
            topic_name,
            self._type,
            error_topic_name=error_topic_name,
            output_topic_name=output_topic_name,
        )

    def add_input_topics(self, topics: list[str]) -> None:
        pass

    def add_extra_input_topic(self, role: str, topics: list[str]) -> None:
        pass

    def set_input_pattern(self, name: str) -> None:
        pass

    def add_extra_input_pattern(self, role: str, topic: str) -> None:
        pass

    def set_output_topic(self, topic_name: str) -> None:
        pass

    def set_error_topic(self, topic_name: str) -> None:
        pass

    def add_extra_output_topic(self, topic_name: str, role: str) -> None:
        pass

    def set_input_topics(self) -> None:
        """
        Puts values of config.from into the streams config section of streams bootstrap
        Supports extra_input_topics (topics by role) or input_topics.
        """
        if self.from_:
            for name, topic in self.from_.topics.items():
                self.apply_from_inputs(name, topic)

    def apply_from_inputs(self, name: str, topic: FromTopic) -> None:
        match (topic.type):
            case InputTopicTypes.INPUT:
                self.add_input_topics([name])
            case InputTopicTypes.EXTRA if topic.role:
                self.add_extra_input_topic(topic.role, [name])
            case InputTopicTypes.INPUT_PATTERN:
                self.set_input_pattern(name)
            case InputTopicTypes.EXTRA_PATTERN if topic.role:
                self.add_extra_input_pattern(topic.role, name)

    def set_output_topics(self) -> None:
        if self.to:
            for name, topic in self.to.topics.items():
                self.apply_to_outputs(name, topic)

    def apply_to_outputs(self, name: str, topic: TopicConfig) -> None:
        match (topic.type):
            case OutputTopicTypes.OUTPUT:
                self.set_output_topic(name)
            case OutputTopicTypes.ERROR:
                self.set_error_topic(name)
            case OutputTopicTypes.EXTRA if topic.role:
                self.add_extra_output_topic(name, topic.role)

    def weave_from_topics(self, prev_component_to: ToSection) -> None:
        """
        Weave output topics of upstream component or from component into config
        Override this method if you want to apply custom logic
        """
        if self.from_:
            return
        input_topics = [
            topic_name
            for topic_name, topic_config in prev_component_to.topics.items()
            if topic_config.type == OutputTopicTypes.OUTPUT
        ]
        if input_topics:
            self.add_input_topics(input_topics)

    def substitute_name(self):
        if self.name:
            self.name = self.substitute_component_names(self.name, self._type)
        else:
            raise ValueError("Every component must have a name in the end.")

    def inflate(self) -> list[PipelineComponent]:
        """Inflate a component. This is helpful if one component should result in multiple components.
        If you wish to support this, override this method and return a list of components the component you result in.
        The order of the components is the order the components will be deployed
        """
        return [self]

    def deploy(self, dry_run: bool) -> None:
        pass

    def destroy(self, dry_run: bool) -> None:
        pass

    def reset(self, dry_run: bool) -> None:
        pass

    def clean(self, dry_run: bool) -> None:
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kpops',
 'kpops.cli',
 'kpops.component_handlers',
 'kpops.component_handlers.helm_wrapper',
 'kpops.component_handlers.kafka_connect',
 'kpops.component_handlers.schema_handler',
 'kpops.component_handlers.topic',
 'kpops.component_handlers.utils',
 'kpops.components',
 'kpops.components.base_components',
 'kpops.components.base_components.models',
 'kpops.components.streams_bootstrap',
 'kpops.components.streams_bootstrap.producer',
 'kpops.components.streams_bootstrap.streams',
 'kpops.pipeline_generator',
 'kpops.utils']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'cachetools>=5.2.0,<6.0.0',
 'dictdiffer>=0.9.0,<0.10.0',
 'pydantic[dotenv]>=1.9.1,<2.0.0',
 'pyhumps>=3.7.3,<4.0.0',
 'python-schema-registry-client>=2.4.1,<3.0.0',
 'requests>=2.28.0,<3.0.0',
 'rich>=12.4.4,<13.0.0',
 'typer[all]>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['kpops = kpops.cli.main:app']}

setup_kwargs = {
    'name': 'kpops',
    'version': '0.6.1',
    'description': 'KPOps is a tool to deploy Kafka pipelines to Kubernetes',
    'long_description': '# KPOps\n\n[![Latest release](https://img.shields.io/github/v/release/bakdata/kpops)](https://github.com/bakdata/kpops/releases/latest)\n[![Build status](https://github.com/bakdata/kpops/actions/workflows/ci.yaml/badge.svg)](https://github.com/bakdata/kpops/actions/workflows/ci.yaml)\n\nFor detailed usage and installation instructions, check out\nthe [documentation](https://bakdata.github.io/kpops/latest).\n\n## Features\n\n- **Deploy Kafka apps to Kubernetes**: KPOps allows to deploy consecutive Kafka Streams applications, producers, and Kafka connectors using an easy-to-read and write pipeline definition. \n- **Configure multiple pipelines and steps**: KPOps comes with various abstractions that simplify configuring multiple pipelines and steps within pipelines by sharing configuration between different applications, like producers or streaming applications.\n- **Customize your components**: KPOps comes with multiple base components (Kafka connect, producer, etc.) and allows you to introduce custom components.\n- **Handle your topics and schemas**: KPOps not only creates and deletes your topics but also registers and deletes your schemas.\n- **Manage the lifecycle of your components**: KPOps can deploy, destroy, reset, and clean your defined components from the Kubernetes cluster.\n\n## Install KPOps\n\nKPOps comes as a [PyPI package](https://pypi.org/project/kpops/). \nYou can install it with [pip](https://github.com/pypa/pip):\n\n```shell\npip install kpops\n```\n\n## Documentation\n\n- [What is KPOps?](https://bakdata.github.io/kpops/latest/user)\n- [Getting started with KPOps](https://bakdata.github.io/kpops/latest/user/getting-started/)\n- [Examples](https://bakdata.github.io/kpops/latest/user/examples)\n\n## Contributing\n\nWe are happy if you want to contribute to this project.\nIf you find any bugs or have suggestions for improvements, please open an issue.\nWe are also happy to accept your PRs.\nJust open an issue beforehand and let us know what you want to do and why.\n\n## License\n\nKPOps is licensed under the [MIT License](https://github.com/bakdata/kpops/blob/main/LICENSE).\n',
    'author': 'bakdata',
    'author_email': 'opensource@bakdata.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/bakdata/kpops',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING, Iterator, Optional

import typer

from kpops.cli.custom_formatter import CustomFormatter
from kpops.cli.pipeline_config import ENV_PREFIX, PipelineConfig
from kpops.cli.registry import Registry
from kpops.component_handlers import ComponentHandlers
from kpops.component_handlers.kafka_connect.connector_handler import ConnectorHandler
from kpops.component_handlers.schema_handler.schema_handler import SchemaHandler
from kpops.component_handlers.topic.handler import TopicHandler
from kpops.component_handlers.topic.proxy_wrapper import ProxyWrapper
from kpops.pipeline_generator.pipeline import Pipeline

if TYPE_CHECKING:
    from kpops.components.base_components import PipelineComponent


LOG_DIVIDER = "#" * 100

app = typer.Typer(
    pretty_exceptions_show_locals=False,
)

BASE_DIR_PATH_OPTION: Path = typer.Option(
    default=Path("."),
    exists=True,
    dir_okay=True,
    file_okay=False,
    envvar=f"{ENV_PREFIX}PIPELINE_BASE_DIR",
    help="Base directory to the pipelines (default is current working directory)",
)

DEFAULT_PATH_OPTION: Path = typer.Option(
    default=Path("defaults"),
    exists=True,
    dir_okay=True,
    file_okay=False,
    envvar=f"{ENV_PREFIX}DEFAULT_PATH",
    help="Path to defaults folder",
)

CONFIG_PATH_OPTION: Path = typer.Option(
    default=Path("config.yaml"),
    exists=False,
    dir_okay=False,
    file_okay=True,
    readable=True,
    envvar=f"{ENV_PREFIX}CONFIG_PATH",
    help="Path to the config.yaml file",
)

PIPELINE_PATH_ARG: Path = typer.Argument(
    default=...,
    exists=True,
    file_okay=True,
    dir_okay=False,
    readable=True,
    envvar=f"{ENV_PREFIX}PIPELINE_PATH",
    help="Path to YAML with pipeline definition",
)

PIPELINE_STEPS: str | None = typer.Option(
    default=None,
    envvar=f"{ENV_PREFIX}PIPELINE_STEPS",
    help="Comma separated list of steps to apply the command on",
)

DRY_RUN: bool = typer.Option(
    True,
    "--dry-run/--execute",
    help="Whether to dry run the command or execute it",
)

COMPONENTS_MODULES: str | None = typer.Argument(
    default=None,
    help="Custom Python module containing your project-specific components",
)

logger = logging.getLogger()
stream_handler = logging.StreamHandler()
stream_handler.setFormatter(CustomFormatter())
logger.addHandler(stream_handler)

log = logging.getLogger("")


def setup_pipeline(
    pipeline_base_dir: Path,
    pipeline_path: Path,
    components_module: str | None,
    pipeline_config: PipelineConfig,
) -> Pipeline:
    registry = Registry()
    if components_module:
        registry.find_components(components_module)
    registry.find_components("kpops.components")

    handlers = setup_handlers(components_module, pipeline_config)
    return Pipeline.load_from_yaml(
        pipeline_base_dir, pipeline_path, registry, pipeline_config, handlers
    )


def setup_handlers(
    components_module: str | None, config: PipelineConfig
) -> ComponentHandlers:
    schema_handler = SchemaHandler.load_schema_handler(components_module, config)
    connector_handler = ConnectorHandler.from_pipeline_config(config)
    proxy_wrapper = ProxyWrapper(config)
    topic_handler = TopicHandler(proxy_wrapper)

    return ComponentHandlers(schema_handler, connector_handler, topic_handler)


def setup_logging_level(verbose: bool):
    logging.getLogger().setLevel(logging.DEBUG if verbose else logging.INFO)


def parse_steps(steps: str) -> set[str]:
    return set(steps.split(","))


def remove_pipeline_prefix(name: str, pipeline: Pipeline) -> str:
    return name.removeprefix(pipeline.config.pipeline_prefix)


def get_step_names(
    steps_to_apply: list[PipelineComponent], pipeline: Pipeline
) -> list[str]:
    return [remove_pipeline_prefix(step.name, pipeline) for step in steps_to_apply]


def filter_steps_to_apply(
    pipeline: Pipeline, steps: set[str]
) -> list[PipelineComponent]:
    skipped_steps: list[str] = []

    def filter_component(component: PipelineComponent) -> bool:
        if remove_pipeline_prefix(component.name, pipeline) in steps:
            return True
        skipped_steps.append(remove_pipeline_prefix(component.name, pipeline))
        return False

    steps_to_apply = list(filter(filter_component, pipeline))
    log.info("KPOPS_PIPELINE_STEPS is defined.")
    log.info(
        f"Executing only on the following steps: {get_step_names(steps_to_apply, pipeline)}"
        f", \n ignoring {skipped_steps}"
    )
    return steps_to_apply


def get_steps_to_apply(
    pipeline: Pipeline, steps: str | None
) -> list[PipelineComponent]:
    return (
        list(pipeline)
        if not steps or steps == '""'  # workaround to allow "" as empty value for CI
        else filter_steps_to_apply(pipeline, steps=parse_steps(steps))
    )


def reverse_pipeline_steps(pipeline, steps) -> Iterator[PipelineComponent]:
    return reversed(get_steps_to_apply(pipeline, steps))


def log_action(action: str, pipeline_component: PipelineComponent):
    log.info("\n")
    log.info(LOG_DIVIDER)
    log.info(f"{action} {pipeline_component.name}")
    log.info(LOG_DIVIDER)
    log.info("\n")


def create_pipeline_config(
    config: Path, defaults: Path, verbose: bool
) -> PipelineConfig:
    setup_logging_level(verbose)
    PipelineConfig.Config.config_path = config
    pipeline_config = PipelineConfig(defaults_path=defaults)
    return pipeline_config


@app.command(
    help="Enriches pipelines steps with defaults. The output is used as input for the deploy/destroy/... commands."
)
def generate(
    pipeline_base_dir: Path = BASE_DIR_PATH_OPTION,
    pipeline_path: Path = PIPELINE_PATH_ARG,
    components_module: Optional[str] = COMPONENTS_MODULES,
    defaults: Path = DEFAULT_PATH_OPTION,
    config: Path = CONFIG_PATH_OPTION,
    print_yaml: bool = typer.Option(
        True, help="Print enriched pipeline yaml definition"
    ),
    save: bool = typer.Option(False, help="Save pipeline to yaml file"),
    out_path: Optional[Path] = typer.Option(
        None, help="Path to file the yaml pipeline should be saved to"
    ),
    verbose: bool = typer.Option(False, help="Enable verbose printing"),
):
    pipeline_config = create_pipeline_config(config, defaults, verbose)
    pipeline = setup_pipeline(
        pipeline_base_dir, pipeline_path, components_module, pipeline_config
    )
    if print_yaml:
        pipeline.print_yaml()

    if save:
        if not out_path:
            raise ValueError(
                "Please provide a output path if you want to save the generated deployment."
            )
        pipeline.save(out_path)


@app.command(help="Deploy pipeline steps")
def deploy(
    pipeline_base_dir: Path = BASE_DIR_PATH_OPTION,
    pipeline_path: Path = PIPELINE_PATH_ARG,
    components_module: Optional[str] = COMPONENTS_MODULES,
    defaults: Path = DEFAULT_PATH_OPTION,
    config: Path = CONFIG_PATH_OPTION,
    verbose: bool = False,
    dry_run: bool = DRY_RUN,
    steps: Optional[str] = PIPELINE_STEPS,
):
    pipeline_config = create_pipeline_config(config, defaults, verbose)
    pipeline = setup_pipeline(
        pipeline_base_dir, pipeline_path, components_module, pipeline_config
    )

    steps_to_apply = get_steps_to_apply(pipeline, steps)
    for component in steps_to_apply:
        log_action("Deploy", component)
        component.deploy(dry_run)


@app.command(help="Destroy pipeline steps")
def destroy(
    pipeline_base_dir: Path = BASE_DIR_PATH_OPTION,
    pipeline_path: Path = PIPELINE_PATH_ARG,
    components_module: Optional[str] = COMPONENTS_MODULES,
    defaults: Path = DEFAULT_PATH_OPTION,
    config: Path = CONFIG_PATH_OPTION,
    steps: Optional[str] = PIPELINE_STEPS,
    dry_run: bool = DRY_RUN,
    verbose: bool = False,
):
    pipeline_config = create_pipeline_config(config, defaults, verbose)
    pipeline = setup_pipeline(
        pipeline_base_dir, pipeline_path, components_module, pipeline_config
    )
    pipeline_steps = reverse_pipeline_steps(pipeline, steps)
    for component in pipeline_steps:
        log_action("Destroy", component)
        component.destroy(dry_run)


@app.command(help="Reset pipeline steps")
def reset(
    pipeline_base_dir: Path = BASE_DIR_PATH_OPTION,
    pipeline_path: Path = PIPELINE_PATH_ARG,
    components_module: Optional[str] = COMPONENTS_MODULES,
    defaults: Path = DEFAULT_PATH_OPTION,
    config: Path = CONFIG_PATH_OPTION,
    steps: Optional[str] = PIPELINE_STEPS,
    dry_run: bool = DRY_RUN,
    verbose: bool = False,
):
    pipeline_config = create_pipeline_config(config, defaults, verbose)
    pipeline = setup_pipeline(
        pipeline_base_dir, pipeline_path, components_module, pipeline_config
    )
    pipeline_steps = reverse_pipeline_steps(pipeline, steps)
    for component in pipeline_steps:
        log_action("Reset", component)
        component.destroy(dry_run)
        component.reset(dry_run)


@app.command(help="Clean pipeline steps")
def clean(
    pipeline_base_dir: Path = BASE_DIR_PATH_OPTION,
    pipeline_path: Path = PIPELINE_PATH_ARG,
    components_module: Optional[str] = COMPONENTS_MODULES,
    defaults: Path = DEFAULT_PATH_OPTION,
    config: Path = CONFIG_PATH_OPTION,
    steps: Optional[str] = PIPELINE_STEPS,
    dry_run: bool = DRY_RUN,
    verbose: bool = False,
):
    pipeline_config = create_pipeline_config(config, defaults, verbose)
    pipeline = setup_pipeline(
        pipeline_base_dir, pipeline_path, components_module, pipeline_config
    )
    pipeline_steps = reverse_pipeline_steps(pipeline, steps)
    for component in pipeline_steps:
        log_action("Clean", component)
        component.destroy(dry_run)
        component.clean(dry_run)


if __name__ == "__main__":
    app()



from enum import Enum


class AppType(Enum):
    STREAMS_APP = "streams-app"
    PRODUCER_APP = "producer-app"
    CLEANUP_STREAMS_APP = "streams-app-cleanup-job"
    CLEANUP_PRODUCER_APP = "producer-app-cleanup-job"



from __future__ import annotations

import importlib
import inspect
import os
import sys
from collections.abc import Iterator
from dataclasses import dataclass, field
from typing import TypeVar

from kpops.cli.exception import ClassNotFoundError
from kpops.components.base_components.pipeline_component import PipelineComponent

T = TypeVar("T")
ClassDict = dict[str, type[T]]  # type -> class

sys.path.append(os.getcwd())


@dataclass
class Registry:
    """Class Registry to store and retrieve PipelineComponents."""

    _classes: ClassDict[PipelineComponent] = field(default_factory=dict, init=False)

    def find_components(self, module_name: str) -> None:
        """
        Find all PipelineComponent subclasses in module
        :param module_name: name of the python module
        """
        for _class in _find_classes(module_name, PipelineComponent):
            self._classes[_class._type] = _class

    def __getitem__(self, component_type: str) -> type[PipelineComponent]:
        try:
            return self._classes[component_type]
        except KeyError:
            raise ClassNotFoundError(
                f"Could not find a component of type {component_type}"
            )


def find_class(module_name: str, baseclass: type[T]) -> type[T]:
    try:
        return next(_find_classes(module_name, baseclass))
    except StopIteration:
        raise ClassNotFoundError


def _find_classes(module_name: str, baseclass: type[T]) -> Iterator[type[T]]:
    module = importlib.import_module(module_name)
    for _, _class in inspect.getmembers(module, inspect.isclass):
        if issubclass(_class, baseclass) and module_name in _class.__module__:
            yield _class



from __future__ import annotations

import logging
import sys
from functools import cached_property
from typing import TYPE_CHECKING

from kpops.component_handlers.helm_wrapper.helm import Helm
from kpops.component_handlers.helm_wrapper.helm_diff import HelmDiff
from kpops.component_handlers.helm_wrapper.model import (
    HelmConfig,
    HelmUpgradeInstallFlags,
)
from kpops.component_handlers.helm_wrapper.utils import trim_release_name
from kpops.component_handlers.kafka_connect.connect_wrapper import ConnectWrapper
from kpops.component_handlers.kafka_connect.exception import ConnectorNotFoundException
from kpops.component_handlers.kafka_connect.model import (
    KafkaConnectConfig,
    KafkaConnectorType,
    KafkaConnectResetterConfig,
    KafkaConnectResetterValues,
)
from kpops.component_handlers.kafka_connect.timeout import timeout
from kpops.utils.colorify import greenify, magentaify, yellowify
from kpops.utils.dict_differ import render_diff

if TYPE_CHECKING:
    from kpops.cli.pipeline_config import KafkaConnectResetterHelmConfig, PipelineConfig

log = logging.getLogger("KafkaConnect")


class ConnectorHandler:
    def __init__(
        self,
        connect_wrapper: ConnectWrapper,
        timeout: int,
        kafka_connect_resetter_helm_config: KafkaConnectResetterHelmConfig,
        broker: str,
        helm_diff: HelmDiff,
        helm_config: HelmConfig = HelmConfig(),
    ):
        self._connect_wrapper = connect_wrapper
        self._timeout = timeout
        self._helm_config = helm_config
        self._helm_repo_config = kafka_connect_resetter_helm_config.helm_config
        self.helm_diff = helm_diff

        helm_repo_config = kafka_connect_resetter_helm_config.helm_config
        self.kafka_connect_resseter_chart = (
            f"{helm_repo_config.repository_name}/kafka-connect-resetter"
        )
        self.chart_version = kafka_connect_resetter_helm_config.version
        self.namespace = (
            kafka_connect_resetter_helm_config.namespace  # namespace where the re-setter jobs should be deployed to
        )
        self.broker = broker
        self.values = kafka_connect_resetter_helm_config.helm_values

    @cached_property
    def helm(self):
        helm = Helm(self._helm_config)
        helm.add_repo(
            self._helm_repo_config.repository_name,
            self._helm_repo_config.url,
            self._helm_repo_config.repo_auth_flags,
        )
        return helm

    def create_connector(
        self,
        connector_name: str,
        kafka_connect_config: KafkaConnectConfig,
        dry_run: bool,
    ) -> None:
        """
        Creates a connector. If the connector exists the config of that connector gets updated.
        :param connector_name: The connector name.
        :param kafka_connect_config: The connector config.
        :param dry_run: If the connector creation should be run in dry run mode.
        """
        if dry_run:
            self.__dry_run_connector_creation(connector_name, kafka_connect_config)
        else:
            try:
                timeout(
                    lambda: self._connect_wrapper.get_connector(connector_name),
                    secs=self._timeout,
                )

                timeout(
                    lambda: self._connect_wrapper.update_connector_config(
                        connector_name, kafka_connect_config
                    ),
                    secs=self._timeout,
                )

            except ConnectorNotFoundException:
                timeout(
                    lambda: self._connect_wrapper.create_connector(
                        connector_name, kafka_connect_config
                    ),
                    secs=self._timeout,
                )

    def destroy_connector(self, connector_name: str, dry_run: bool) -> None:
        """
        Deletes a connector resource from the cluster.
        :param connector_name: The connector name.
        :param dry_run: If the connector deletion should be run in dry run mode.
        """
        if dry_run:
            self.__dry_run_connector_deletion(connector_name)
        else:
            try:
                timeout(
                    lambda: self._connect_wrapper.get_connector(connector_name),
                    secs=self._timeout,
                )

                timeout(
                    lambda: self._connect_wrapper.delete_connector(connector_name),
                    secs=self._timeout,
                )
            except ConnectorNotFoundException:
                log.warning(
                    f"Connector Destruction: the connector {connector_name} does not exist. Skipping."
                )

    def clean_connector(
        self,
        connector_name: str,
        connector_type: KafkaConnectorType,
        dry_run: bool,
        retain_clean_jobs: bool,
        **kwargs,
    ) -> None:
        """
        Cleans the connector from the cluster. At first, it deletes the previous cleanup job (connector resetter)
        to make sure that there is no running clean job in the cluster. Then it releases a cleanup job.
        If the retain_clean_jobs flag is set to false the cleanup job will be deleted.
        :param connector_name: Name of the connector
        :param connector_type: Type of the connector (SINK or SOURCE)
        :param dry_run: If the cleanup should be run in dry run mode or not
        :param retain_clean_jobs: If the cleanup job should be kept
        :param kwargs: Other values for the KafkaConnectResetter
        """
        suffix = "-clean"
        clean_up_release_name = connector_name + suffix
        trimmed_name = trim_release_name(clean_up_release_name, suffix)

        log.info(
            magentaify(
                f"Connector Cleanup: uninstalling cleanup job Helm release from previous runs for {connector_name}"
            )
        )
        self.__uninstall_kafka_resetter(trimmed_name, dry_run)

        log.info(
            magentaify(
                f"Connector Cleanup: deploy Connect {connector_type.value} resetter for {connector_name}"
            )
        )

        stdout = self.__install_kafka_resetter(
            trimmed_name, connector_name, connector_type, dry_run, kwargs
        )

        if dry_run and self.helm_diff.config.enable:
            current_release = self.helm.get_manifest(trimmed_name, self.namespace)
            new_release = Helm.load_helm_manifest(stdout)
            helm_diff = HelmDiff.get_diff(current_release, new_release)
            self.helm_diff.log_helm_diff(helm_diff, log)

        if not retain_clean_jobs:
            log.info(magentaify("Connector Cleanup: uninstall Kafka Resetter."))
            self.__uninstall_kafka_resetter(trimmed_name, dry_run)

    def __dry_run_connector_creation(
        self, connector_name: str, kafka_connect_config: KafkaConnectConfig
    ) -> None:
        try:
            connector_config = self._connect_wrapper.get_connector(connector_name)

            log.info(
                yellowify(
                    f"Connector Creation: connector {connector_name} already exists."
                )
            )
            if diff := render_diff(
                connector_config.config, kafka_connect_config.dict()
            ):
                log.info(f"Updating config:\n {diff}")

            log.debug(kafka_connect_config.dict(exclude_unset=True, exclude_none=True))
            log.debug(f"PUT /connectors/{connector_name}/config HTTP/1.1")
            log.debug(f"HOST: {self._connect_wrapper.host}")
        except ConnectorNotFoundException:
            log.info(
                greenify(
                    f"Connector Creation: connector {connector_name} does not exist. Creating connector with config:\n {kafka_connect_config}"
                )
            )
            log.debug("POST /connectors HTTP/1.1")
            log.debug(f"HOST: {self._connect_wrapper.host}")

        errors = self._connect_wrapper.validate_connector_config(kafka_connect_config)
        if len(errors) > 0:
            log.error(
                f"Connector Creation: validating the connector config for connector {connector_name} resulted in the following errors:"
            )
            log.error("\n".join(errors))
            sys.exit(1)
        else:
            log.info(
                greenify(
                    f"Connector Creation: connector config for {connector_name} is valid!"
                )
            )

    def __dry_run_connector_deletion(self, connector_name: str) -> None:
        try:
            self._connect_wrapper.get_connector(connector_name)
            log.info(
                magentaify(
                    f"Connector Destruction: connector {connector_name} already exists. Deleting connector."
                )
            )
            log.debug(f"DELETE /connectors/{connector_name} HTTP/1.1")
            log.debug(f"HOST: {self._connect_wrapper.host}")
        except ConnectorNotFoundException:
            log.warning(
                f"Connector Destruction: connector {connector_name} does not exist and cannot be deleted. Skipping."
            )

    def __install_kafka_resetter(
        self,
        release_name: str,
        connector_name: str,
        connector_type: KafkaConnectorType,
        dry_run: bool,
        kwargs,
    ) -> str:
        return self.helm.upgrade_install(
            release_name=release_name,
            namespace=self.namespace,
            chart=self.kafka_connect_resseter_chart,
            dry_run=dry_run,
            flags=HelmUpgradeInstallFlags(
                version=self.chart_version, wait_for_jobs=True, wait=True
            ),
            values={
                **KafkaConnectResetterValues(
                    config=KafkaConnectResetterConfig(
                        connector=connector_name,
                        brokers=self.broker,
                        **kwargs,
                    ),
                    connector_type=connector_type.value,
                    name_override=connector_name,
                ).dict(),
                **self.values,
            },
        )

    def __uninstall_kafka_resetter(self, release_name: str, dry_run: bool) -> None:
        self.helm.uninstall(
            namespace=self.namespace,
            release_name=release_name,
            dry_run=dry_run,
        )

    @classmethod
    def from_pipeline_config(
        cls, pipeline_config: PipelineConfig
    ) -> ConnectorHandler:  # TODO: annotate as typing.Self once mypy supports it
        return cls(
            connect_wrapper=ConnectWrapper(host=pipeline_config.kafka_connect_host),
            timeout=pipeline_config.timeout,
            kafka_connect_resetter_helm_config=pipeline_config.kafka_connect_resetter_config,
            helm_config=pipeline_config.helm_config,
            broker=pipeline_config.broker,
            helm_diff=HelmDiff(pipeline_config.helm_diff_config),
        )



from enum import Enum

from pydantic import BaseConfig, BaseModel, Extra


class TopicSpec(BaseModel):
    topic_name: str
    partitions_count: int | None
    replication_factor: int | None
    configs: list[dict[str, str]] | None

    class ModelConfigs(BaseConfig):
        extra = Extra.forbid


class TopicResponse(BaseModel):
    kind: str = "KafkaTopic"
    metadata: dict[str, str]
    cluster_id: str
    topic_name: str
    is_internal: bool
    replication_factor: int
    partitions: dict[str, str]
    configs: dict[str, str]
    partition_reassignments: dict[str, str]
    partitions_count: int = (
        1
        # HACK: workaround for field being absent in Kafka REST response
        # https://github.com/confluentinc/kafka-rest/issues/1085
    )

    class ModelConfigs(BaseConfig):
        extra = Extra.forbid


class KafkaTopicConfigSource(str, Enum):
    DYNAMIC_TOPIC_CONFIG = "DYNAMIC_TOPIC_CONFIG"
    DEFAULT_CONFIG = "DEFAULT_CONFIG"


class KafkaTopicConfigSynonyms(BaseModel):
    name: str
    value: str
    source: KafkaTopicConfigSource

    class Config(BaseConfig):
        extra = Extra.allow


class KafkaTopicConfig(BaseModel):
    source: KafkaTopicConfigSource
    synonyms: list[KafkaTopicConfigSynonyms]
    value: str
    name: str

    class Config(BaseConfig):
        extra = Extra.allow


class TopicConfigResponse(BaseModel):
    data: list[KafkaTopicConfig]

    class Config(BaseConfig):
        extra = Extra.allow


class KafkaBrokerConfigSource(str, Enum):
    STATIC_BROKER_CONFIG = "STATIC_BROKER_CONFIG"
    DYNAMIC_BROKER_CONFIG = "DYNAMIC_BROKER_CONFIG"
    DEFAULT_CONFIG = "DEFAULT_CONFIG"


class KafkaBrokerConfigSynonyms(BaseModel):
    name: str
    value: str | None
    source: KafkaBrokerConfigSource

    class Config(BaseConfig):
        extra = Extra.allow


class KafkaBrokerConfig(BaseModel):
    source: KafkaBrokerConfigSource
    synonyms: list[KafkaBrokerConfigSynonyms]
    value: str | None
    name: str

    class Config(BaseConfig):
        extra = Extra.allow


class BrokerConfigResponse(BaseModel):
    data: list[KafkaBrokerConfig]

    class Config(BaseConfig):
        extra = Extra.allow



import logging
from typing import Iterable

from kpops.component_handlers.helm_wrapper.model import HelmDiffConfig, HelmTemplate
from kpops.utils.dict_differ import Change, render_diff

log = logging.getLogger("HelmDiff")


class HelmDiff:
    def __init__(self, config: HelmDiffConfig):
        self.config = config

    @staticmethod
    def get_diff(
        current_release: Iterable[HelmTemplate],
        new_release: Iterable[HelmTemplate],
    ) -> list[Change]:
        new_release_index = {
            helm_template.filepath: helm_template for helm_template in new_release
        }

        changes: list[Change] = []
        # collect changed & deleted files
        for current_resource in current_release:
            # get corresponding dry-run release
            new_resource = new_release_index.pop(current_resource.filepath, None)
            if new_resource:
                change = Change(
                    old_value=current_resource.template, new_value=new_resource.template
                )
            else:
                change = Change(old_value=current_resource.template, new_value={})
            changes.append(change)

        # collect added files
        for new_resource in new_release_index.values():
            changes.append(Change(old_value={}, new_value=new_resource.template))

        return changes

    def log_helm_diff(self, changes: list[Change], logger: logging.Logger):
        for change in changes:
            if change.old_value and change.new_value:
                if diff := render_diff(
                    change.old_value,
                    change.new_value,
                    ignore=self.config.ignore,
                ):
                    logger.info("\n" + diff)
